open Pervasives2

(* B: en attendant un peu mieux.... *)
open Transitions
open Term
open Definitions

type statelist= Statelist.t

type automaton =
    { trans: Transitions.t;
      mutable fstates: Stateset.t
    }
		
let make : unit -> automaton = fun _ ->
	{trans = Transitions.make () ; fstates = Stateset.empty;}

let transtable (a: automaton) : Transitions.transtable =
  a.trans.transitions

let state_is_final (a: automaton) (t: term) = match t with
 | State q -> Stateset.mem q a.fstates
	| _ -> raise Not_a_state

(* Mapping a state to the corresponding term.
   Necessary to produce relevant outputs when detecting patterns (associated to a state
   in automaton representing lhs and patterns) *)

type s2t = (state, term) Hashtbl.t

(* mapping a state to the corresponding variable for lhs automaton *)
type statevar = (state, variable) Hashtbl.t

let state_list_to_string l =
  String.concat " " (List.map State.to_string l)

(* Tree automaton copy *)
let copy (a: automaton) : automaton =
  { trans = Transitions.copy a.trans;
    fstates = a.fstates
  }


(* Build a tree automaton from an external_trans list and a set of final states *)

let external_to_TA (fs: Stateset.t)(ls: Transitions.external_trans list): automaton=
  let max = (List2.fmax (function (Trans (_, _, q)) -> q | (Epsilon (q1,q2,_)) -> max q1 q2) 0 ls) + 1 in
  let transT = Transitions.make () in
    (* We build states and tables accordingly to the maximum state label *)
    (*for i=0 to max do ignore(Transitions.make_state transT) done;*)
		Transitions.make_n_states transT (max + 1);
    (try 
      List.iter (function (Trans t) -> Transitions.add transT t 
                   | (Epsilon (q,q',phi)) -> if not(phi=Formula.mk_false) then Transitions.addR transT (q,q',phi) else ()) ls
    with Invalid_argument("index out of bounds") -> failwith "Bad state max number!");
    { fstates = fs; trans = transT }


let new_statevar () = Hashtbl.create Common.max_var

let new_s2t () = Hashtbl.create Common.max_equation

let statevar_add (s: statevar) (q1: state) (v: variable) : unit =
  try
    if Hashtbl.find s q1 <> v then
      failwith
        "Automaton:statevar_add: different variable association for a state"
  with
    | Not_found -> Hashtbl.add s q1 v

(** == [Hashtbl.find] *)
let statevar_find = Hashtbl.find

(** membership of a state in a statevar *)
let statevar_mem= Hashtbl.mem

(** get the list of states associated to variables *)
let statevar_to_list (t:statevar)= Hashtbl.fold (fun i j l -> i::l) t []

let statevar_to_string t =
  Hashtbl.fold
    (fun k e str ->
       (State.to_string k) ^ " -> " ^ (Variable.to_string e) ^ "\n" ^ str)
    t ""

let s2t_to_string t =
  Hashtbl.fold
    (fun k e str ->
       (State.to_string k) ^ " -> " ^ (Term.to_string e) ^ "\n" ^ str)
    t ""

(* Run function guided by destination state.
   If t=f(q1) dest=q2 and delta contains
   f(q1) -> q3 and q3 -> q2 it is OK.  (The Transitions.get_trans is modulo epsilon)

   If t=f(q1) and dest=q2, then it is covered by a delta where
   f(q3) -> q2 and q1 ->E q3.

   If t=f(q1) and dest=q2, then it is NOT covered by a delta where
   f(q3) -> q2 and q1 ->R q3.
*)

(* answers true if it is possible to rewrite t to dest with transitions and
   epsilon transitions. Does not rewrite states of t *)

let rec oriented_run (d: Transitions.t) (t: term) (dest: state) : bool =
  Verbose.print 3
    (fun () -> "Oriented Run for t=" ^ (Term.to_string t) ^
       "\nwith delta=" ^ (Transitions.to_string d) ^ "\n");

  let oriented_subrun (tl: term list) (destl: state list) : bool =
    try
      List.for_all2 (oriented_run d) tl destl
    with
      | Invalid_argument _ ->
          failwith "Automaton:oriented_run_list: size of lists is not the same!"
  in
    match t with
      | Fsym (f, tl) ->
	  List.exists
	    (fun (_, sl, _) -> oriented_subrun tl sl)
	    (Transitions.get_trans dest f d)
      | State q ->
          q = dest || Transitions.is_succE d q dest (* q=dest or q =E dest *)
      | Var _ ->
          failwith "Automaton:oriented_run: non ground term!"


exception Configuration_not_recognised of term
exception Automaton_is_not_deterministic of term

(* Deterministic run of t, with no epsilon resolving *)
let rec det_run (d: Transitions.t) : term -> term = function
  | Fsym (f, []) ->
      (match Transitions.get_trans_sym_constant f d with
 	 | [] -> raise (Configuration_not_recognised (Fsym (f, []))) (* B: le cas ou il n'y a aucune transition *)
	 | [(_, _, q)] -> State q 
   | _ -> raise (Automaton_is_not_deterministic  (Fsym (f, [])))                              
      )
  | Fsym (f, tl) ->
      (* B: reduction des sous-termes par delta *)
      let res_tl = List.map (det_run d) tl in
        (*(try*)
	   let res_states = List.map Term.get_state res_tl in
             (* B: calcul assez obscure des transitions de la forme
                f(q_1, \dots, q_n) -> q, en fait je crois qu'on recupere directement
                la liste de la forme [(q_1^1, \dots, q_n^1); \dots; (q_1^m, \dots, q_n^m)]
                ou chaque tuple j pour 1 <= j <= m, est de la forme [q_1^j; \dots; q_n^j]
             *)
	   let relevant_trans =
	     List.flatten
               (List.map
                  (fun q -> Transitions.get_trans_sym_parameter f q d)
		  res_states
               )
           in
	       	(match List.find_all (fun  (_, sl ,_) -> sl = res_states) relevant_trans with
        		| [] -> raise (Configuration_not_recognised (Fsym (f, res_tl)))
				 		| [_, _, q] ->   State q
						| _ -> raise (Automaton_is_not_deterministic (Fsym(f, res_tl)))
						)

	 (*with (* Transition is not fully normalizable *)
	   | Term.Not_a_state -> Fsym (f, res_tl)
               (* B: qui balance cette exception et pourquoi ???
                  ca doit etre Term.get_state mais a quoi sert cette derniere
                  on n'a pas deja des etats ?!?!?
                  Reponse: get_state (State q) -> q sinon Not_a_state
               *)
        )*)
  | State q -> State q
  | Var _ -> failwith "Automaton:oriented_run: non ground term!"



(* Non deterministic run: use any transition to normalize a term (with epsilonE resolving only).
   Used for application of equations: given a \sigma and an equation l=r produces, for l\sigma all
   possible states reachable, modulo epsilonE. *)
(* gives every possible state in which a term can be rewritten or the empty list *)
(* Principle is to find all transitions rooted by the same top symbol (all
   f(q1...qn)->q for f(t1...tn) and then removing parameter states q...qn of transitions
   one after one, and keep only transitions such that they were matching t1...tn successively *)

(* if l\sigma -> q and q=q' then, by default, we don't give q' as a "reachable" state, since it will be added
   as a successor by E of all new predecessors of q by transitivity closure of epsilon E. More precisely,
   if r\sigma q'', then we only give the result q for l\sigma. Since all_run is only used for equation application
   we have l\sigma -> q and r\sigma -> q'' and q <->q'. By adding q <->q'' we add (by transitivity) q'' <-> q'.
   Hence, no need to give q' as a result.
 *)
let rec all_run (d: Transitions.t) (t: term) : state list =
  (*  let _= print_string ("Term= "^(Term.to_string t)^"\n") in *)
  match t with
    | Fsym (f, []) ->
        (match Transitions.get_trans_sym_constant f d with
	   | [] -> []
	   | l -> List.map (fun  (_, _, q) -> q) l
        )
    | Fsym (f, tl) ->
        all_run_list d tl (Transitions.simple_get_trans_f f d)
    | State q -> [q]
    | Var "_" ->
        failwith "Automaton:all_run: anonymous variable cannot be used on top!"
    | Var _ ->
        failwith "Automaton:all_run: free variable uninstantiated!"
          (* all_run_list recursively run through all transitions of l
             and either remove the 1st parameter of each transition if
             it is accepted, or reject the complete transition *)
          (* B: Reprendre ici !!! *)
          (* La fonction all_run_list execute tous les run possible
             sur les sous-termes en parallele *)

(* print_string ("Term list= "^(String.concat ", " (List.map Term.to_string tl))^
   "\non l=" ^ (Transitions.symextlist_to_string l)^"\n"); *)
(** @param l est typiquement la liste de toutes les transitions de d ayant un certain symbole comme tête, par exemple f
@param tl est typiquement la liste des sous-termes sous le symbole f dans le terme dont all_run calcule les descendants
@return une liste d'états qui sera renvoyée par all_run, donc ce doit être la liste des états en lesquels f(tl) est reconnu *)
and all_run_list (d: Transitions.t) (tl: term list) (l: external_symtrans list) : state list =
  let remain_trans =
    List.fold_left
      (fun trans subterm -> (* external_symtrans list -> term -> external_symtrans list *)
         match subterm with (* 2 premiers cas sont des hacks inutiles à la compréhension de l'algo*)
           | Var "_" -> List.map (fun (f, sl, dest) -> f, List.tl sl, dest) trans
               (* anonymous variables can match any state i.e. keep all
                  trans, only remove the 1st parameter of every trans *)
           | Var x -> failwith ("Automaton:all_run: non ground term: variable " ^ x)
               (* | State q -> Common.list_omap (fun (f, sl, dest) ->
                  let q' = List.hd sl in if q' = q ||
                  (Transitions.is_predE q q' d) then Some (f, List.tl
                  sl, dest) else None ) trans *)
               (* B: Ce cas-la n'est pas vraiment utile, il est deja
                  pris en compte dans le suivant...  Raison
                  d'optimalite ??? *)
           | State(_) | Fsym(_,_) as t -> (* cas général intéressant *)
               let states = all_run d t in (* liste des états en lesquels subterm est reconnu *)
                 List2.optmap (* effet global du bazar sur la liste ltrans :
								pour chaque « transition » f(q1,…,qn) → dest, on regarde si q1 [mod E] reconnait subterm 
								si tel n'est pas le cas, la transition est éliminée
								si tel est bien le cas, on ampute la transition de q1 pour en faire la pseudo-transition f(q2,…,qn) → dest
								Cette amputation du premier état permet de conserver à chaque étape du fold_left la propriété que
								subterm = t_i, notant tl=[t_1,…,t_k], et le premier état de chaque pseudo-transition de ltrans est le q_i de la vraie transition dont elle est issue*)
	           (fun (f, sl, dest) -> (* trans -> state option *) (* en fait, ce n'est pas forcément une vraie transition, à cause des amputations*) 
                      let q' = List.hd sl in (* sl ne peut pas être vide : au départ, les listes d'états dans les transitions de l ont toutes la même longueur que tl ; à chaque étape du fold_left, on avance dans tl et on décapite les listes d'états des transitions ; à la fin on vide complètement ces listes, mais alors le fold_left est fini *)
                        (* if q' is an Esucc of one state q of sl then it
                           is ok, i.e.  t -> Delta+E q ->E q' *)
		        if
                          List.exists
                            (fun q -> q = q' || (Transitions.is_succE d q' q)) (* égalité mod E *)
                            states
                        then (* dans states, il y a un état égal [mod E] à l'état en tête de sl
													ie. l'état en tête de sl est parmi [mod E] ceux en lesquels subterm est reconnu *)
                          Some (f, List.tl sl, dest) (* Achtung ! ceci ressemble à une transition, mais n'en n'est plus vraiment une car le nombre de sous termes sous t est incorrect*)
		        else (* sinon : on jette cette transition*)
                          None
                   ) trans
      ) l tl
  in (* on n'a gardé que les transitions f(q1,…,qn)→dest dont les qi sont atteignables depuis t_i : on récupère les dest *)
    List.map (fun (_, _, dest) -> dest) remain_trans (*noter qu'à ce stade, les « transitions » sont salement amochées : la liste d'états a été vidée *)


(*   Normalization adds transitions to delta *)

let rec normalize_term (d: Transitions.t) (t: term) (wvar: bool): (state * external_trans list)=
  match t with
    | Fsym (f, []) ->
	let relevant_trans = Transitions.get_trans_sym_constant f d in
	  (match relevant_trans with
	     | [] ->
		 let s = Transitions.make_state d in
		   Transitions.add d (f, [], s);
		   (s, [Trans(f, [], s)])
	     | (_,_,q)::_ -> (q, [])
          )
    | Fsym (f, tl) ->
	let (res_state_list, translist) = normalize_term_list d tl wvar in
	  (try
	     let relevant_trans =
               List.fold_right
                 (fun q acc -> Transitions.get_trans_sym_parameter f q d @ acc)
		 res_state_list
                 []
             in
	       try
		 let _, _, q =
                   List.find (fun (_, sl, _) -> sl = res_state_list) relevant_trans
		 in
                   (q, translist)
                     (* A transition matches, no new state is necessary *)
               with (* No transition matches, a new state and a new transition are necessary *)
		 | Not_found ->
		     let s = Transitions.make_state d in
		       Transitions.add d (f, res_state_list, s);
		       (s, (Trans (f, res_state_list, s)) :: translist)
	   with (* Transition is not fully normalizable *)
	     | Term.Not_a_state ->
                 failwith "Automaton:normalize_term: Subterms are not fully normalized!"
          )
    | State q ->
	if wvar then
          failwith "Automaton:normalize_term: cannot be a state"
	else (q, [])
    | Var x ->
	if wvar then
	  (* For variables, in the case where we normalize a lhs
	     we simply add a new constant (and state) if it is
	     not already present *)
	  let relevant_trans = Transitions.get_trans_sym_constant x d in
	    match relevant_trans with
	      | [] ->
		  let s = Transitions.make_state d in
		    Transitions.add d (x,[],s);
		    (s, [Trans(x, [], s)])
	      | (_, _, q)::_ -> (q, [])
	else
          failwith "Automaton:normalize_term: cannot be a variable!"

and normalize_term_list (d: Transitions.t) (terms: term list) (wvar: bool): state list * external_trans list =
  (* let res = List.fold_right
     (fun t acc ->
     let q, trans = normalize_term d t wvar in
     let sl, trans' = acc in
     (q::sl, trans @ trans')
     )
     terms ([], [])
     in*)
  Verbose.print 0 (fun _ -> "Warning: B: normalize_term_list (let res) vs normalize_term_list2 ne produisent pas la meme chose\n");
  normalize_term_list2 d terms wvar
    (* DBG: pour le warning ci-dessus
       if res = normalize_term_list2 d terms wvar then
       res
       else
       failwith "le fold_right et l'autre fonction ne calcule pas la meme chose..."
    *)
and normalize_term_list2 d terms wvar =
  (* B: Version originale de la fonction, c'est a dire sans le bug !!!! *)
  match terms with
    | [t] -> let (s, tl) = normalize_term d t wvar in
	([s], tl)
    | t::rem1 ->
	let (s, tl) = normalize_term d t wvar in
	let (sl2, tl2) = normalize_term_list d rem1 wvar in
	  (s::sl2,  tl @tl2)
    | [] -> failwith "Automaton:normalize_term_list: empty list!"



(* Returns a list such that its first element is the transition f(sl)->q, for
   any q, if such a transition exists in Delta. Otherwise, it searches for
   transitions f(sl')->q', for any q', such that for all state qi of sl there
   exists a state qi' in sl' (at the same rank) such that qi ->epsiE* qi'.  For
   simplicity and efficiency, if f(sl) -> q does not occur, but f(sl')->q'
   exists, we give back the first transition found and not all of them.

   YB : Attention : il me semble qu'il y a une erreur ici.
   j'ai l'impression que l'on considère comme atome (q,q') à partir du moment où nous avons q ->*E q'.
   ce qui n'est plus le cas.
*)

let closest_transitions (f: symbol) (sl: state list) (d: Transitions.t) : (external_symtrans * Formula.t) list =
  (* We check if there exists a q such that f(sl) -> q *)
  let _ = IFDEF TEST THEN
    Transitions.print_fclosure d.epsilonR
    ENDIF
  in
  let relevant_trans =
    List.flatten (List.map (fun q -> Transitions.get_trans_sym_parameter f q d) sl)
  in
    try
      (* If such a transition exists, we return it *)
      [ List.find (fun (_, sl2, _) -> sl = sl2) relevant_trans, Formula.mk_true ]
    with
      | Not_found ->
          (* Otherwise we need to check modulo epsilonE on the SAME set of transition *)
          try
            let (_, sl2, q) =
              List.find
                (fun (_, sl2, _) ->
                   try
                     List.for_all2 (fun s1 s2 -> s1 = s2 || is_succE d s1 s2) sl sl2
                   with Invalid_argument _ -> assert false
                ) relevant_trans
            in
            let phi =
              List.fold_right2
                (fun s1 s2 f -> if s1 = s2 then f else
	           let interesting = d.epsilonE#fclosure.(s2).(s1) in
	             Formula.mk_and interesting f
                ) sl sl2 Formula.mk_true
            in
              [(f, sl2, q), phi]
          with
            | Not_found ->
                (* Otherwise we check on transitions that are related
                   to states of sl by epsilonE *)
                (* we first gather all successors of states of sl *)
                let succ = List.fold_right (fun s sl -> (get_succE d s) @ sl) sl [] in
                let relevant_trans =
                  List.flatten
                    (List.map (fun q -> Transitions.get_trans_sym_parameter f q d) succ)
                in
                  try
                    let (_, sl2, q) =
                      List.find
                        (fun (_, sl2, _) ->
                           try
                             List.for_all2 (fun s1 s2 -> s1 = s2 || is_succE d s1 s2) sl sl2
                           with Invalid_argument _ -> assert false
                        ) relevant_trans
                    in
                    let phi =
                      List.fold_right2
                        (fun s1 s2 f -> if s1 = s2 then f else
                           let interesting = d.epsilonE#fclosure.(s2).(s1) in
	 	             Formula.mk_and interesting f
                        ) sl sl2 Formula.mk_true
                    in
                      [ (f, sl2, q), phi ]
	          with
                    | Not_found -> []




(* Normalize_termE adds transitions to a transition set.

   If normalization is impossible using Delta alone, we try using
   Delta ∪ epsilonE ...  and return the formula corresponding to the
   epsilonE transitions used. Such normalization cannot be applied to
   terms with variables.

   The result is a top state, a formula and the corresponding transitions.

   For simplicity, in this implementation, we only keep track of the
   first solution (except for the case where the term is a state
   (i.e. we need all solutions).  *)

let rec normalize_termE (t: term) (d: Transitions.t) :
    (state * Formula.t * external_trans list) =
  match t with
    | Fsym (f, []) ->
	let relevant_trans = Transitions.get_trans_sym_constant f d in
	  (match relevant_trans with
	     | [] ->
		 let s = Transitions.make_state d in
		   Transitions.add d (f, [], s);
		   (s, Formula.mk_true, [Trans(f, [], s)])
	     | (_, _, q)::_ -> (q, Formula.mk_true, [])
          )
    | Fsym (f, tl) ->
        (* We normalize subterms of f: tl *)
	let (res_state_list,phi,translist) = normalize_termE_list tl d in
	   (* relevant_trans is the list of the transitions closest (in term of
              number of applications of epsiE necessary) to res_state_list *)
	 let relevant_trans = closest_transitions f res_state_list d in
(*         let _ = print_string ("Closest trans for "^(Symbol.to_string f)^"("^(state_list_to_string res_state_list)^") is: "^
                               (String.concat "\n" 
                                  (List.map 
                                     (function (t,phi) -> ("("^(Transitions.symext_to_string t)^","^(Formula.to_string phi)^")")) relevant_trans))^"\n") in
*)
           (match relevant_trans with
              | [] -> (* No transition matches, a new state and a new transition are necessary *)
		  let s = Transitions.make_state d in
		    Transitions.add d (f, res_state_list, s);
		    (s, phi, (Trans (f, res_state_list, s)) :: translist)
              | ((_, sl, s), phi2) :: _ ->
                  (s, Formula.mk_and phi phi2, translist)
	   )
    | State q -> (q, Formula.mk_true, [])
    | Var x ->
        failwith "No variable in terms for normalize_termE"

and normalize_termE_list (tl: term list) (d: Transitions.t) :
    (state list * Formula.t * external_trans list) =
  match tl with
    | [t] -> let (s, phi, tl) = normalize_termE t d in
	([s], phi, tl)
    | t::rem1 ->
	let (s, phi1, tl) = normalize_termE t d in
	let (sl2, phi2, tl2) = normalize_termE_list rem1 d in
	  (s::sl2, Formula.mk_and phi1 phi2, List.append tl tl2)
    | [] -> failwith "Automaton:normalize_term_list: empty list!"



(* Normalisation of a term for adding in delta, during completion (flag variable false) *)
(* Also: normalisation of a term to produce a term_automaton: for left-hand
   side of rules, or initial term sets (flag variable true) *)

(* Add a term to a delta and returns the top state of this term
   as well as the list of new external transitions resulting in
   this adding. Uses d to normalize t and adds transitions for t to d. Also
   creates new states in d when necessary.

   The boolean wvar is a flag which is true when normalizing terms with
   variables (left-hand side of rules) *)


(* Normalisation of a term for adding in delta, using only new states! *)

let rec normalize_new_states (t: term) (d: Transitions.t) : state * external_trans list =
  match t with
    | Fsym (f, []) ->
	let s = Transitions.make_state d in
	  Transitions.add d (f,[],s);
	  (s, [Trans(f,[],s)])
    | Fsym (f, tl) ->
	let (res_state_list, translist) = normalize_new_states_list tl d in
	let s = Transitions.make_state d in
	  Transitions.add d (f, res_state_list,s);
	  (s, (Trans (f, res_state_list, s)) :: translist)
    | State q -> (q, [])
    | Var x -> failwith "Automaton:normalize_term: cannot be a variable!"

and normalize_new_states_list (tl: term list) (d: Transitions.t) :
state list * external_trans list =
    match tl with
      | [t] -> let (s, tl)= normalize_new_states t d in
	  ([s], tl)
      | t::rem1 ->
	  let (s, tl)= normalize_new_states t d in
	  let (sl2, tl2)= normalize_new_states_list rem1 d in
	    (s::sl2, List.append tl tl2)
      | [] -> failwith "Automaton:normalize_term_list: empty list!"



(* Normalisation of a term for adding in delta, during completion (flag variable
   false. This corresponds to oldstyle normalization, i.e. normalize f(t1...tn)
   ->q such that the top transition is of the form f(q1...qn) -> q.

   Normalization adds transitions to delta.
 *)

let oldnormalize_term (newstates: bool)(t: term)(dest: state)(d: Transitions.t) : external_trans list=
  match t with
    | Fsym (f, []) ->
	if Transitions.add' d (f, [], dest) then
	  [ Trans (f, [], dest) ]
	else
          []

    | Fsym (f, tl) ->
	let (res_state_list, translist) =
	  if newstates then
            normalize_new_states_list tl d
	  else
            normalize_term_list d tl false
	in
	  if Transitions.add' d (f,res_state_list,dest) then
	    Trans(f,res_state_list,dest)::translist
	  else
            translist
    | State q ->
        [ Epsilon (q, dest, Formula.mk_true) ]
    | Var x ->
        failwith "Automaton:oldnormalize_term: cannot be a variable!"

(*   Normalization adds transitions to delta and associates the states to their recognized term in stt *)

let rec normalize_lhs (d: Transitions.t) (t: term) (wvar: bool) (stt: s2t): (state * external_trans list)=
  match t with
    | Fsym (f, []) ->
	let relevant_trans = Transitions.get_trans_sym_constant f d in
	  (match relevant_trans with
	     | [] ->
		 let s = Transitions.make_state d in
		   Transitions.add d (f, [], s);
                   Hashtbl.add stt s (Fsym (f,[]));
		   (s, [Trans(f, [], s)])
	     | (_,_,q)::_ -> (q, [])
          )
    | Fsym (f, tl) ->
	let (res_state_list, translist) = normalize_lhs_list d tl wvar stt in
	  (try
	     let relevant_trans =
               List.fold_right
                 (fun q acc -> Transitions.get_trans_sym_parameter f q d @ acc)
		 res_state_list
                 []
             in
	       try
		 let _, _, q =
                   List.find (fun (_, sl, _) -> sl = res_state_list) relevant_trans
		 in
                   (q, translist)
                     (* A transition matches, no new state is necessary *)
               with (* No transition matches, a new state and a new transition are necessary *)
		 | Not_found ->
		     let s = Transitions.make_state d in
		       Transitions.add d (f, res_state_list, s);
                       Hashtbl.add stt s (Fsym (f,tl));
		       (s, (Trans (f, res_state_list, s)) :: translist)
	   with (* Transition is not fully normalizable *)
	     | Term.Not_a_state ->
                 failwith "Automaton:normalize_lhs: Subterms are not fully normalized!"
          )
    | State q ->
	if wvar then
          failwith "Automaton:normalize_lhs: cannot be a state"
	else (q, [])
    | Var x ->
	if wvar then
	  (* For variables, in the case where we normalize a lhs
	     we simply add a new constant (and state) if it is
	     not already present *)
	  let relevant_trans = Transitions.get_trans_sym_constant x d in
	    match relevant_trans with
	      | [] ->
		  let s = Transitions.make_state d in
		    Transitions.add d (x,[],s);
                  (* We do not add states in stt *)
		    (s, [Trans(x, [], s)])
	      | (_, _, q)::_ -> (q, [])
	else
          failwith "Automaton:normalize_lhs: cannot be a variable!"

and normalize_lhs_list (d: Transitions.t) (terms: term list) (wvar: bool) (stt: s2t): state list * external_trans list =
  match terms with
    | [t] -> let (s, tl) = normalize_lhs d t wvar stt in
	([s], tl)
    | t::rem1 ->
	let (s, tl) = normalize_lhs d t wvar stt in
	let (sl2, tl2) = normalize_lhs_list d rem1 wvar stt in
	  (s::sl2,  tl @tl2)
    | [] -> failwith "Automaton:normalize_lhs_list: empty list!"


(* utilisé pour construire alhs *)

let term_to_automaton (tl: term list) (vl: variable list) (stt: s2t) : automaton * statevar * state list =
  let rec aux (tl: term list) (a: automaton) (var_assoc: statevar) (fs: state list) :
      automaton * statevar * state list =
    match tl with
      | [] -> (a, var_assoc, List.rev fs)
      | t::rem ->
	  let (final, trans) = normalize_lhs a.trans t true stt in
	    if (not(Hashtbl.mem stt final)) then Hashtbl.add stt final t;
	    List.iter
              (function
		 | Trans (f,_,dest) ->
                     if List.mem f vl then statevar_add var_assoc dest f
		 | _ ->
                     failwith "Automaton.term_to_automaton: transitions cannot be epsilon here!"
              ) trans;
	    aux rem { trans = a.trans;
                      fstates = Stateset.add final a.fstates
                    } var_assoc (final::fs)
  in
    aux tl { trans = Transitions.make ();
             fstates = Stateset.empty
           } (new_statevar ()) []


let to_string (a: automaton) : string =
  ("States "^(state_list_to_string (Transitions.get_all_states a.trans))^"\nFinal States "^
     (state_list_to_string (Stateset.elements a.fstates))^"\nTransitions\n"^(Transitions.to_string a.trans))

let to_timbuk (a: automaton) : string =
  ("States "^(state_list_to_string (Transitions.get_all_states a.trans))^"\nFinal States "^
     (state_list_to_string (Stateset.elements a.fstates))^"\nTransitions\n"^(Transitions.to_timbuk a.trans))



(* Merges all symtables of l into acc *)

let rec merge_symtable (l: (symbol * table) list) (acc: symtable): unit =
  match l with
    | [] -> ()
    | (f, t)::rem ->
	(try
           let table = Hashtbl.find acc f in
	     Hashtbl.replace acc f (Statetable.union t table)
	 with Not_found -> Hashtbl.add acc f t
        );
        merge_symtable rem acc


(* Checks if there is at least one common symbol between list l1 and l2 *)

let rec matchone_symbol (l1: (symbol * table) list) (t2: symtable) =
  match l1 with
    | [] -> false
    | (f1, _) :: rem ->
	Hashtbl.mem t2 f1 || (matchone_symbol rem t2)


(* Filters a list of (symbol->table) s.t. the first list contains only
   constants and the second one only non constant transitions *)

let rec get_const (l: (symbol * table) list) (resconst: (symbol * table) list) (resnotconst: (symbol * table) list): (symbol * table) list * (symbol * table) list =
  match l with
    | [] -> (resconst, resnotconst)
    | (f, t)::rem ->
        if Transitions.constant_table t then
          get_const rem ((f, t)::resconst) resnotconst
        else
          get_const rem resconst ((f, t)::resnotconst)


(* Check if there is at least one common constant symbol between l1 and t2 *)
(* Symbols of l1 are only constants *)

let rec fast_const_test (l1: (symbol*table)list)(t2: symtable): bool=
  match l1 with
      [] -> false
    | (s,t)::rem ->
	(Hashtbl.mem t2 s)||(fast_const_test rem t2)

(* Ordering couples so that the smaller becomes the first one *)

let couple_order s1 s2=
  if compare s1 s2 >0 then (s2,s1)
  else (s1,s2)

(* Check or store the fact that the intersection of two states is not empty *)
let mem_couple s1 s2 t: bool=
  let (c1,c2)= couple_order s1 s2 in
    try (Stateset.mem c2 (Hashtbl.find t c1))
    with Not_found -> false

let add_couple s1 s2 t =
  let (c1, c2) = couple_order s1 s2 in
    try
      let setc1 = Hashtbl.find t c1 in
        Hashtbl.replace t c1 (Stateset.add c2 setc1)
    with Not_found ->
      Hashtbl.add t c1 (Stateset.singleton c2)


(* Get or store the FORMULA associated to the fact that the intersection of two states is not empty *)

let mem_couple_formula s1 s2 t: Formula.t=
  let (c1,c2)= couple_order s1 s2 in
    try (Hashtbl.find t (c1,c2))
    with Not_found -> Formula.mk_false

let add_couple_formula s1 s2 f t: unit =
  if Formula.unsat f then ()
  else
    let (c1, c2) = couple_order s1 s2 in
    try
      let fex = Hashtbl.find t (c1,c2) in
      (* If the existing formula is "weaker" than the new one we replace, i.e.
         A /\ B is stored and we found A (A/\B => A) ... or A is stored and we found True (A => True). *)
      if Formula.subsume fex f then
        Hashtbl.replace t (c1,c2) f        
      else ()
    with Not_found -> Hashtbl.add t (c1,c2) f



(* test of non emptyness of intersection of a list of couple of states
   of a single tree automaton  ** modulo epsilon ** *)
(* input list [(q1,q2);(q3,q4)] checks if q1/\q2 is not empty AND
  q3/\q4 not empty *)
(* Terminates also on cycling empty tree automata *)
(* Returns the formula to satisfy for the intersection not to be empty:
   - True: in this case we have a proof that whatever the approximation may be, the intersection 
   is not empty.
   - Phi different from True or False, then we know that when Phi is true the approximation is not empty...
     There may be a solution for Phi=True but it may not be already found.
   - False: the intersection is empty whatever the approximation may be. *)

let not_empty_intersection (l: (state * state) list)(delta: Transitions.t): Formula.t=
  (* table for inductive couple of states (for intersection of
     recursive transitions *)
  let inductive = Hashtbl.create Common.max_state_symb in
  (* [ok] table for saving formulas for each couple of states (s1,s2) having a non empty intersection *)
  let ok = Hashtbl.create Common.max_state_symb in
  let deltatable = delta.Transitions.transitions in
  (* Check that intersection between list of states lq1 and lq2 is not empty
     The result is the formula necessary to satisfy so that the intersection is not empty *)
  let rec check_state_list (lq1: state list) (lq2: state list) (acc: Formula.t): Formula.t =
    match lq1, lq2 with
      | [], [] -> acc
      | q1::rem1, q2::rem2 -> 
          let f= (check_state q1 q2) in
          Verbose.printf 3 "Check_state_list: Intersec: (q%i,q%i): %s\n" q1 q2 (Formula.to_string f);
          if Formula.unsat f then f
          else check_state_list rem1 rem2 (Formula.mk_and f acc)
      | _, _ ->
          failwith "Automaton:not_empty_intersection: state lists should have the same length"

  (* Check that intersection between q1 and q2 is not empty.
     The result is the formula necessary to satisfy so that the intersection is not empty *)
  and check_state q1 q2: Formula.t=
    let f= mem_couple_formula q1 q2 ok in
    if not(Formula.unsat f) then f
    else
      if mem_couple q1 q2 inductive then
      (* Couple has already been considered we are now dealing with
	 recursive transitions which are of no help for checking emptyness of an
	   intersection*)
      Formula.mk_false
        (* Otherwise, it has to be investigated more deeply *)
    else 
        if q1 = q2 then 
          begin
            add_couple_formula q1 q2 Formula.mk_true ok;
            Formula.mk_true
          end
        else
          let f1= (Transitions.is_succRE_formula delta q1 q2) in
          if Formula.is_sat f1 then
            begin
              add_couple_formula q1 q2 f1 ok;
              f1
            end
          else
            let f2= (Transitions.is_succRE_formula delta q2 q1) in
            if Formula.is_sat f2 then
              begin
                add_couple_formula q1 q2 f2 ok;
                f2
              end
            else
      (* We are considering all predecessors of q1 and q2 in order to
	 cover epsilon closure. *)
              let allq1 = (Transitions.get_predRE delta q1) in
              let allq2 = (Transitions.get_predRE delta q2) in
              let tableq1 = (Hashtbl.copy (Hashtbl.find deltatable q1)), Formula.mk_true in
              let tableq2 = (Hashtbl.copy (Hashtbl.find deltatable q2)), Formula.mk_true in
              let ltables1 =
                tableq1::
                  (List.map
	             (fun (q,phi) -> ((Hashtbl.find deltatable q)),phi)
	             allq1
                  )
              in
              let ltables2 =
                tableq2::
                  (List.map
	             (fun (q,phi) -> ((Hashtbl.find deltatable q)),phi)
	             allq2
                  )
              in
      let _ = add_couple q1 q2 inductive in
      let f= check_intersection_table_list ltables1 ltables2 in
      if Formula.is_sat f then add_couple_formula q1 q2 f ok;
      f

  (* Check that there is at least one common term recognized by a symbol table (seen as a list) [l1] and 
     table t2 *)
  and check_transition_list (l1: (symbol * table) list)(t2: symtable) : Formula.t =
    match l1 with
      | [] -> Formula.mk_false
      | (f, t) :: rem ->
	  try
            (let tab2 = Hashtbl.find t2 f in
             let t1= Statetable.elements t in
             let st2= Statetable.elements tab2 in
             let f=check_first_transition t1 st2 in
             if Formula.is_sat f then f
             else check_transition_list rem t2)
          with Not_found -> check_transition_list rem t2

  (* Check that there is at least statelist in list [l1] whose intersection is not empty w.r.t. statelist list [l2] *)
  and check_first_transition (l1: statelist list) (l2: statelist list): Formula.t=
    match l1 with 
        [] -> Formula.mk_false
      | sl::rem -> let f= check_first_transition2 sl l2 in
                   if Formula.is_sat f then f
                   else check_first_transition rem l2

  and check_first_transition2 (s1: statelist) (l2: statelist list): Formula.t=
    match l2 with 
        [] -> Formula.mk_false
      | fsl::rem -> 
          let f= check_state_list s1 fsl Formula.mk_true in
          if Formula.is_sat f then f
          else (check_first_transition2 s1 rem)

(* Check intersection of lists of sym tables. Results in the formula for the first transitions to match.
   not necessary the simplest formula (True) but the first *)

  and check_intersection_table_list (lt1: (symtable * Formula.t) list) (lt2: (symtable * Formula.t) list): Formula.t=
  let rec check_one_table (stconst: (symbol * table) list) (stnonconst: (symbol * table) list)
      (lt2: (symtable * Formula.t) list): Formula.t=
    match lt2 with
        [] -> Formula.mk_false
      | (stable,phi)::rem -> 
        (* Checks if there is at least one common symbol between list
           l1 and l2 for efficiency reasons *)
          if (matchone_symbol stconst stable) || (matchone_symbol stnonconst stable) 
          then
            (* there is at least one common symbol *)
            if (fast_const_test stconst stable) 
            (* If there is a common constant, we are done *)
            then phi
            else 
              let formula_trans_list= check_transition_list stnonconst stable in
              if Formula.is_sat formula_trans_list 
              then Formula.mk_and phi formula_trans_list
              else check_one_table stconst stnonconst rem 
          else
            check_one_table stconst stnonconst rem
  in
  match lt1 with 
      [] -> Formula.mk_false
    | (stable,phi)::rem -> 
          let (const, nonconst) = get_const (Hashtbl2.elements stable) [] [] in
          let f= (check_one_table const nonconst lt2) in
          if Formula.is_sat f then Formula.mk_and phi f
          else check_intersection_table_list rem lt2
  in
  let (lq1, lq2) = List.split l in
    check_state_list lq1 lq2 Formula.mk_true


(* utilise par inter_of' pour indiquer l'etat de traitement des couples d'etats *)
type visited = Open | Closed | Done
(* un couple d'etats est soit
   - Open : en cours de visite
   - Done : on a calcule la formule
   - Closed : on ne connait pas sa formule
              et n'est pas en cours de traitement
*)
(* Une deuxieme version sans sauvegarde des resultats... *)
let inter_of' (a1: automaton) (a2: automaton) : Formula.t =
  (* maxi is the maximal label number of a state in ai *)
  let last = a1.trans.epsilonR#last and last' = a2.trans.epsilonR#last in
    (* Les tuples (q, q', Φ) qui dénotent l'intersection a₁ × a₂
       aux états q et q', avec Φ la condition de a₁ nécessaire pour satisfaire
       la non-vacuité en (q, q'). c.f. le papier
       De plus, pour l'implémentation ne respecte pas l'algo du papier,
       le calcul est orienté des états finaux vers les co-accessibles,
       à l'inverse du papier... Normalement plus efficace, car il ne considère que
       les couples d'états co-accessibles.
    *)
  let visited = Array.make_matrix (last + 1) (last' + 1) Closed
  and phi = Array.make_matrix (last + 1) (last' + 1) Formula.mk_false
  and alpha = a1.trans.epsilonR#fclosure in

  (* Building formula for couple of states (q,q') exploring epsilonRE predecessors of q *)
  let rec phi_of_states q q' =
    Verbose.printf 3 "phi_of_states: Visiting couple of states (q%i, q%i')\n" q q';
    match visited.(q).(q') with
      | Open -> Formula.mk_false (* (q, q') deja  en cours d'exploration... *)
      | Done -> phi.(q).(q')
      | Closed ->
          (try
             (* For all predecessor q1 of q, we call phi_of_fun *)
             for q1 = 0 to last do
               Verbose.println 3 (fun _ -> "phi for "^(State.to_string q1)^" ->RE "^(State.to_string q)^"||"^(Formula.to_string alpha.(q).(q1))^"\n");
               if Formula.is_sat alpha.(q).(q1) then (* q₁ —α––>^*_RE q *)
                 let phi_q1_q' = match visited.(q1).(q') with
                   | Open -> Formula.mk_false
                   | Done -> phi.(q1).(q')
                   | Closed ->
                       visited.(q1).(q') <- Open;
                       let phi_q1_q' = phi_of_fun q1 q' in
                         visited.(q1).(q') <-
                           if Formula.tautology phi_q1_q' then Done else Closed;
                         phi_q1_q'
                 in
(*                 let _= print_string ("phi("^(State.to_string q)^")("^(State.to_string q')^"')= "^(Formula.to_string phi.(q).(q'))^"\n") in *)
                 phi.(q1).(q') <- Formula.mk_or phi.(q1).(q') phi_q1_q';
                 (* if not(Formula.subsume phi_q1_q' phi.(q1).(q')) then phi.(q1).(q') <- Formula.mk_or phi.(q1).(q') phi_q1_q'; *)
                 phi.(q).(q') <- Formula.mk_or phi.(q).(q')
                     (Formula.mk_and alpha.(q).(q1) phi_q1_q');
(*                 let _= print_string ("new phi("^(State.to_string q)^")("^(State.to_string q')^"')= "^(Formula.to_string phi.(q).(q'))^"\n") in *)
                   if Formula.tautology phi.(q).(q') then raise Break;
             done
           with Break -> phi.(q).(q') <- Formula.mk_true
          );
          Verbose.print 3 (fun _ -> "  Φ("^(State.to_string q)^", "^(State.to_string q')^")= "^(Formula.to_string phi.(q).(q'))^"\n");
          visited.(q).(q') <- Done;
          phi.(q).(q')

  (* Building formula for couple of states (q,q') exploring
     transitions f(...)->q and f(...)->q' *)
  and phi_of_fun q q' =
    Verbose.printf 3 "phi_of_fun: Visiting couple of states (q%i, q%i')\n" q q';
    try
      let st  = Hashtbl2.find a1.trans.transitions q
      and st' = Hashtbl2.find a2.trans.transitions q' in
        Verbose.print 3 (fun () -> "st= "^(Hashtbl2.to_string st));
        Verbose.print 3 (fun () -> "\nst'= "^(Hashtbl2.to_string st')^"\n");
      let phi_q_q' = ref Formula.mk_false in
        Hashtbl2.iter
          (fun f largs' ->
             try Statetable.iter
               (fun args -> Statetable.iter
                  (* we found a transition f(q1,...,qn) ->q and f(q'1, ..., q'n) ->q'
                     args= (q1,...,qn) and args'=(q'1, ..., q'n)*)
                  (fun args' ->
                     let alpha =
                       try (* α = /\₁ⁿ Φ_i pour [(q₁, q₁'); ...; (q_n, q'_n)] *)
                         List.fold_left2
                           (fun f qi qi' -> let phi_qi_qi' = phi_of_states qi qi' in
                              if not (Formula.is_sat phi_qi_qi') then raise Break;
                              Formula.mk_and phi_qi_qi' f)
                           Formula.mk_true args args'
                       with Break -> Formula.mk_false
                     in
                       phi_q_q' := Formula.mk_or !phi_q_q' alpha;
                       if Formula.tautology !phi_q_q' then raise Break;
                       (* phi_q_q' = ⊤ plus la peine de continuer *)
                  ) largs'
               ) (Hashtbl2.find st f)
             with Not_found -> () (* pas de transition de la forme q <- f(...) *)
          ) st';
        !phi_q_q'
    with
      | Break -> Formula.mk_true (* on a trouve ⊤ *)
      | Not_found -> Formula.mk_false (* st ou st' est vide donc il n'y a pas de transitions
                                         q <- f(...) ou pas de transition q' <- g(...)
                                         pour tout f ou g *)
  in
    (* We start from the couples (q,q') where q is a final state of a1 and q' a final state of q *)
    Stateset.fold
      (fun qf phi_inter ->
         Stateset.fold
         (fun qf' phi_inter ->
            Formula.mk_or (phi_of_states qf qf') phi_inter
         ) a2.fstates phi_inter
      ) a1.fstates Formula.mk_false

(* Peut etre utile pour debugguer le code de inter_of
   let str_of pref =
   List2.to_string ~left:(f ^ "(") ~right:")" ~sep:", "
   (fun q -> pref ^ (string_of_int q))
   in
   Format.printf "%s %s |- Φ = %s\n@?"
   (str_of "q" args) (str_of "q'" args') (Formula.to_string alpha);
*)




(*  0 => a raffiné;
 *  1 => pas de collision entre a et bad;
 * -1 => il existe un contre-example *)
let refine a bad =
  let phi = ref (inter_of' a bad) in
  let _= Verbose.println 3 (fun _ -> "Formula resulting of the intersection : "^
                              (Formula.to_string !phi)) in
    if not (Formula.is_sat !phi) then    (* A ∩ Bad = ∅ *)
      1                                 (* property holds *)
    else if (Formula.tautology !phi) then
      -1                                (* there exists a counter-example *)
    else (                              (* Refinement *)
      while Formula.is_sat !phi do
        Verbose.println 2 (fun _ -> "Refining Step!");
        Verbose.println 3 (fun _ -> "Refining Step: Φ = " ^ (Formula.to_string !phi));
        Transitions.prune a.trans !phi;
        (* print_string ("\n"^ (Transitions.to_string a.trans));
        print_string "\n"; *)
        phi := inter_of' a bad
      done;
      Verbose.println 3 (fun _ -> "\n" ^ (Transitions.to_string a.trans) ^ "\n");
      0;
    )

let rename (sl: state list) (q: state) (q':state)=
  if (List.mem q' sl) then q else q'

(* Applying a renaming (every state of l is renamed into q) to a list of transitions *)
let rec merge_one ((q,l): state * (state list)) (ltrans: Transitions.external_trans list)(acc: Transitions.external_trans list): Transitions.external_trans list=
  match ltrans with
    | [] -> acc
    | Trans (f,sl,q')::rem ->
        merge_one (q,l) rem
(*          (List2.add (Trans (f,List.map (rename l q) sl, (rename l q q'))) acc) *)
          ((Trans (f,List.map (rename l q) sl, (rename l q q')))::acc)
    | Epsilon (q1,q2,f)::rem ->
        (* renaming of q1 -> q2 may result in q1 -> q1 (or q2 -> q2) however this transition may be
           relevant since it may represent a rewriting step from q1 to q1. *)
        merge_one (q,l) rem
          (* (List2.add (Epsilon ((rename l q q1),(rename l q q2),f)) acc) *)
          ((Epsilon ((rename l q q1),(rename l q q2),f))::acc)

(* Applying one renaming on a state set. (every state of l is renamed into q) *)

let stateset_renaming_one ((q,l): state * (state list)) (set: Stateset.t): Stateset.t=
  List.fold_right (fun s2 set -> Stateset.replace s2 q set) l set

(* Applying a list of renaming (for every state list ls of lfusion, ls is of the
   form q::l2, where all states of l2 are renamed into q) every state of l is
   renamed into q) to a list of transitions *)

let rec merge (lfusion: state list list) (ltrans: Transitions.external_trans list): Transitions.external_trans list=
  match lfusion with
      [] -> ltrans
    | (q::sl)::rem -> merge rem (merge_one (q,sl) ltrans [])
    | _ -> failwith "Automaton.merge: badly formed connex component list"


let stateset_renaming (lfusion: state list list) (set: Stateset.t): Stateset.t=
  List.fold_right 
    (fun renaming set -> match renaming with
         [] | [_] -> failwith "Automaton.stateset_renaming: badly formed connex component list"
       | q::sl -> stateset_renaming_one (q,sl) set) lfusion set

let init_renum_tabl (a: automaton) : (state, state) Hashtbl.t =
	let m = a.trans.next_state - 1 in
	let res = Hashtbl.create m in
	for i = 0 to m do
		Hashtbl.add res i i
	done;
	res


let renumerotation ?table (a: automaton) : ((state, state) Hashtbl.t * automaton) =
  let renum (l: external_trans list) (fs: Stateset.t) : ((state, state) Hashtbl.t * external_trans list * Stateset.t) =
  	let tbl = match table with
			| None -> Hashtbl.create 0
			| Some (t) -> t
		and next = ref 0 in
  	let embed x =
  		try Hashtbl.find tbl x
  		with Not_found -> let nv = !next in incr next; Hashtbl.add tbl x nv; nv
  	in let do_one t = match t with
  		| Trans(f, sl, d) -> Trans(f, List.map embed sl, embed d)
  		| Epsilon(q1, q2, phi) -> Epsilon(embed q1, embed q2, phi)
  	in
  	let res = List.map do_one l
  	and resfs = Stateset.fold (fun x acc -> Stateset.add (embed x) acc) fs Stateset.empty
  	in
  	 (tbl, res, resfs)
		
	and les_trans = Transitions.trans_to_list a.trans.transitions in
	let (tbl, res, resfs) = renum les_trans a.fstates
	in (tbl, external_to_TA resfs res)
																										
(* nettoyage accessibilité et utilité *)
(* algos repris de Timbuk2 *)
let accessibility_cleaning ?hook (a: automaton) : automaton =
	let les_trans = Transitions.trans_to_list a.trans.transitions in
	let rec une_passe (a_traiter : external_trans list) (en_sursis : external_trans list) (bonnes : external_trans list) (et_acc : Stateset.t) (ajout : bool) =
			match a_traiter with
				| [] ->
								if not ajout then
									bonnes, et_acc
								else
									une_passe en_sursis [] bonnes et_acc false
				| Trans(_, sl, dest) as trans :: rem ->
								if List.for_all (fun x -> Stateset.mem x et_acc) sl then
									une_passe rem en_sursis (trans::bonnes) (Stateset.add dest et_acc) true
								else
									une_passe rem (trans::en_sursis) bonnes et_acc ajout
				| Epsilon(q1,q2,_) as trans ::rem  ->
								if Stateset.mem q1 et_acc then
									une_passe rem en_sursis (trans::bonnes) (Stateset.add q2 et_acc) true
								else
									une_passe rem (trans::en_sursis) bonnes et_acc ajout
	in let (oktrans, accessibles) = une_passe les_trans [] [] Stateset.empty false
	in let nfinal = Stateset.inter accessibles a.fstates
	in (match hook with
		| None -> ()
		| Some (hk) -> hk accessibles);
		external_to_TA nfinal oktrans

let utility_cleaning ?hook (a: automaton) : automaton =
	let les_trans = Transitions.trans_to_list a.trans.transitions in
	let rec une_passe (a_traiter : external_trans list) (en_sursis : external_trans list) (bonnes : external_trans list) (et_ut : Stateset.t) =
		match a_traiter with
			| [] ->
				bonnes, et_ut
			| Trans(_, sl, dest) as trans :: rem ->
				if Stateset.mem dest et_ut then
					une_passe (en_sursis @ rem) [] (trans::bonnes) (Stateset.add_list sl et_ut)
				else
					une_passe rem (trans::en_sursis) bonnes et_ut
			| Epsilon(q1,q2,_) as trans :: rem ->
				if Stateset.mem q2 et_ut then
					une_passe (en_sursis @ rem) [] (trans::bonnes) (Stateset.add q1 et_ut)
				else
					une_passe rem (trans::en_sursis) bonnes et_ut

	in let (trans_utiles, etats_utiles) = une_passe les_trans [] [] a.fstates
  in (match hook with
  		| None -> ()
  		| Some (hk) -> hk etats_utiles);	
	external_to_TA a.fstates trans_utiles

(* 
   From an automaton A having epsilonR and epsilonE transitions, this function exports
   an epsilon-free tree automaton A' such that L(A)=L(A') and every possible run in A 
   is possible in A'.
   This is necessary for preserving R-closedness of tree automata.

   Export a tree automaton with epsilonE and epsilonR transitions to a tree
   automaton without epsilon transitions. We use mergin of cycles on epsilonR
   AND epsilonE, i.e. if q1 -> q2 -> q3 -> q1 then states q1, q2, and q3 are
   merged, i.e. renamed. Otherwise for all transition q1 -> q2 every transition
   c -> q1 is propagated to q2, i.e. we add c -> q2 to the tree
   automaton. Furthermore, we try to preserve R-closedness (of the automaton with
   RE-epsilon) on the tree automaton without epsilon: for all transition
   f(...q...) -> q' if q1 -> q, we add the transition f(...q1...) -> q'...
   otherwise the run f(...q1...) -> q is not possible in epsilon-free automaton 
   and may result in a tree automaton that is not R-closed.
*)


let export (a: automaton)=
  (* Les composantes connexes de <-RE*->. *)
  let cc= Transitions.get_strongly 
            (fun q q' -> (Transitions.is_succRE a.trans q q')&&(Transitions.is_succRE a.trans q' q))
            a.trans in
    Verbose.exec 3 (fun _ -> (Transitions.print_fclosure a.trans.epsilonE));
    Verbose.println 3 (fun _ -> "\nComposantes Connexes= "^
                       (List2.to_string (List2.to_string string_of_int) cc));
    (* We build the list of all states that disappear because of merging *)
    let erased_states= 
      List.fold_right (fun elt l -> 
                         match elt with
                             [] | [_] -> failwith "Bad strongly connected component"
                           | _::r -> List.rev_append r l) cc [] 
    in
      (* Fusion des classes d'équivalence sur toutes les transitions *)
    let all_merged= merge cc (Transitions.to_list a.trans) in
    let (all_sym, all_epsi)= Transitions.partition all_merged in
      (* We propagate epsilon transitions in upward direction, i.e. if f(...q1...) -> q0 and q2 -> q1, 
         we add f(...q2...) -> q0 *)
    let prop_up= Transitions.propagate_epsilon_up all_epsi all_sym in
      (* We propagate epsilon transitions in downward direction, i.e. if f(...q1...) -> q0 and q0 -> q2, 
         we add f(...q1...) -> q2 *)
    let final_prop= Transitions.propagate_epsilon prop_up erased_states a.trans in
      external_to_TA (stateset_renaming cc a.fstates) (List.rev_map (function t -> Trans t) final_prop)



(* Export a tree automaton with epsilonE and epsilonR transitions to a tree automaton with epsilon transitions 
   where all the epsilonE cycles are merged, i.e. if q1 -> q2 -> q3 -> q1 then
   states q1, q2, and q3 are merged, i.e. renamed. *)

let export_epsi (a: automaton) =
  (* Les composantes connexes de <-E*->. E seul! *)
  let cc= Transitions.get_strongly (fun q q' -> (Transitions.is_succE a.trans q q') &&
               (Transitions.is_succE a.trans q' q)) a.trans in
    Verbose.exec 3 (fun _ -> (Transitions.print_fclosure a.trans.epsilonE));
    Verbose.println 3 (fun _ -> "\nComposantes Connexes= "^
                         (List2.to_string (List2.to_string string_of_int) cc));

  (* La propagation des epsilon -RE> sur toutes les transitions de base *)
  let alltrans = Transitions.to_list a.trans in
  let all_merged = merge cc alltrans in
    external_to_TA (stateset_renaming cc a.fstates) all_merged
		
module Produit =
struct
  type produit = {
  	a: automaton;
  	memo_embed : (state * state, state) Hashtbl.t;
  	memo_embedrev : (state, state * state) Hashtbl.t;	
  }

  (** [embed self q p] renvoie l'état de self.a qui correspond à l'état produit (q, p)
  
  si cet état n'avait pas été enregistré, il est créé
  *)
  let embed ?hook (self: produit) (q: state) (p: state) : state =
  	let (qp, ajoute) = (try  (Hashtbl.find self.memo_embed (q, p), false)
  	with Not_found ->
  		let ns = Transitions.make_state self.a.trans in
  			Hashtbl.add self.memo_embed (q,p) ns;
  			Hashtbl.add self.memo_embedrev ns (q,p);
  			(ns, true))
  	in
  		(match hook with
  			| None -> ()
  			| Some(hk) -> hk q p qp ajoute);
  		qp;;
  		
  	
  let embed_list ?hook (self: produit) (ql: state list) (pl: state list) : state list =
  		List.map2 (embed ?hook self) ql pl;;
  
  let embed_rev (self: produit) (qp: state) : state * state =
  	Hashtbl.find self.memo_embedrev qp;;
  
  let proj_q (self: produit) (qp : state) : state =
  	fst (embed_rev self qp);;
  	
  let proj_p (self: produit) (qp: state) : state =
  	snd (embed_rev self qp);;
  
  let is_final (self: produit) (qp: state) : bool =
  	Stateset.mem qp self.a.fstates;;
  
  let qp_to_string self qp =
  			let (q, p) = embed_rev self qp in
  				"<q"^(string_of_int q)^",p"^(string_of_int p)^">";; 
  
  let rec cqp_to_string self cqp =
  	match cqp with
    | Fsym (symbol, []) -> Symbol.to_string symbol
    | Fsym (symbol, l_subterms) ->
        symbol ^ (List2.to_string ~sep:"," ~left:"(" ~right:")" (cqp_to_string self) l_subterms)
  
    | State t -> qp_to_string self t
    | Var x -> x
  
  let prod_imatrix_to_string ?(ghost = false) ?(timbuk = false) (prod: produit) : string =
    let etable = prod.a.trans.Transitions.epsilonE in
  	let str = ref "" and t = etable#degrees
                     and maxind = etable#last in
      for i = 0 to maxind do
        for j = 0 to maxind do
  	match t.(i).(j) with
  	  | Alive x ->
                str := !str ^ (qp_to_string prod j) ^
                  (if timbuk then
                     " → "
                   else
                     " --" ^ (string_of_int x) ^ "--> "
                  ) ^ (qp_to_string prod i) ^ "\n"
            | Ghost ->
                if ghost && not timbuk then
  	        str := !str ^ (qp_to_string prod j) ^ " --G--> " ^
                    (qp_to_string prod i) ^ "\n"
  	  | Away -> ()
        done
      done;
      !str
  
  let prod_delta_to_string (prod : produit) =
  	let t = prod.a.trans.Transitions.transitions
  	in let table_to_string (s: symbol) (dest: state) (t: table) : string =
    if constant_table t then
      (Symbol.to_string s) ^ " → " ^ (qp_to_string prod dest) ^ "\n"
    else
      let str_start = Symbol.to_string s ^ "("
      and str_end = ") → " ^ (qp_to_string prod dest) ^ "\n" in
        List.fold_left
          (fun str sl -> str ^ str_start ^ (List2.to_string ~sep:", " ~left:"" ~right:"" (qp_to_string prod) sl) ^ str_end)
          "" (Statetable.elements t)
  	in
  	List.fold_left
      (fun str (q, symtable) ->
         List.fold_left
           (fun str (f, table) -> str ^ (table_to_string f q table))
           str
           symtable
      )
      ""
      (List.map
         (fun (q, symtable) -> q, Hashtbl2.elements symtable)
         (Hashtbl2.elements t)
      )
  
  let prodtrans_to_string (prod: produit) : string =
  	(prod_delta_to_string prod) ^"\nεE = "^ (prod_imatrix_to_string ~timbuk:true prod)^"\n"

  let access_hook_std ?hook (prod: produit) (accessibles:Stateset.t) : unit =
  	let inaccessibles = List.filter (fun x -> not (Stateset.mem x accessibles)) (Transitions.get_all_states prod.a.trans) in
  	let traite_un qp =
  		let (q, p) = embed_rev prod qp in
  		(match hook with
  			| None -> ()
  			| Some(hk) -> hk q p qp);
  		Hashtbl.remove prod.memo_embed (q, p);
  		Hashtbl.remove prod.memo_embedrev qp
  	in
  		List.iter traite_un inaccessibles;;

  (** crée un automate produit de a0 par a1
  
  il est supposé que les deux automates ne contiennent pas d'ε-transition*)
  let init_produit ?accessibility_hook ?embed_hook ~final_crit (a0: automaton) (a1: automaton) : produit =
  	let resultat = {a=make (); memo_embed = Hashtbl.create 0; memo_embedrev =  Hashtbl.create 0}; 
  	in
  		(* pour chaque symbole f, on va ajouter les transitions produit
  		dont le symbole de tête est f ;
  		au passage, on va créer beaucoup d'états du produit*)			
  		let faire_symbole (f, _) = (* structure (f,_) à cause du type de !Global.alpha *)
  			let ltrans_a0 = simple_get_trans_f f a0.trans (* liste des trans f(…)→… dans a0*)
  			and ltrans_a1 = simple_get_trans_f f a1.trans (* idem dans airr *)
  			and prod_t_t (f0, sl0, dest0) (f1, sl1, dest1) = (* on aura f0=f1*)
  				Transitions.add resultat.a.trans (f0, embed_list ?hook:embed_hook resultat sl0 sl1, embed ?hook:embed_hook resultat dest0 dest1);
  				if (final_crit dest0 dest1) then resultat.a.fstates <- (Stateset.add (embed resultat dest0 dest1) resultat.a.fstates)
  			in let prod_t_la1 trans = (* prend une trans ajoute les trans produit*)
  				List.iter (prod_t_t trans) ltrans_a1
  			in List.iter prod_t_la1 ltrans_a0
  			
  		in List.iter faire_symbole !Global.alpha; (* c'est moche : l'alphabet est stocké dans une ref globale !*)
  		{resultat with a = (accessibility_cleaning ~hook:(access_hook_std ?hook:accessibility_hook resultat) resultat.a)} ;;


end


module Determinisation =
struct
		
		type det_state = Stateset.t
		type det_symtrans = symbol * det_state list * det_state
		
		type representation =  (det_state, state) Hashtbl.t
		type representation_rev = (state, det_state) Hashtbl.t
		
		type det_obj = {
			input : automaton;
			mutable next_state : state;
			memo_repr : representation;
			memo_repr_rev : representation_rev;
			}

		let rp_state (self: det_obj) (ds: det_state) : state =
			let res = (try Hashtbl.find self.memo_repr ds
			with Not_found -> 
					let nouveau = self.next_state in
						self.next_state <- self.next_state + 1;
						Hashtbl.add self.memo_repr ds nouveau;
						Hashtbl.add self.memo_repr_rev nouveau ds;
						nouveau)
			in
				res;;

		let rp_trans (self: det_obj) ((f, dsl, ddest) : det_symtrans) : external_symtrans =
			let sl = List.map (rp_state self) dsl
			and dest = rp_state self ddest
			in (f, sl, dest);;

		let rp_state_rev (self: det_obj) (s: state) : det_state =
			Hashtbl.find self.memo_repr_rev s;;

		let detstate_union (ds1: det_state) (ds2: det_state) : det_state =
			Stateset.union ds1 ds2;;

		let detstate_is_final (self: det_obj) (ds: det_state) : bool =
			Stateset.exists (fun x -> Stateset.mem x self.input.fstates) ds;;

		let is_included (ds1 : det_state) (ds2: det_state) : bool =
			Stateset.for_all (fun x -> Stateset.mem x ds2) ds1;;

		let detsl_included (dsl1: det_state list) (dsl2: det_state list) : bool =
			List.for_all2 (fun ds1 ds2 -> is_included ds1 ds2) dsl1 dsl2;;
		
		let get_containing (ds: det_state) (ld: det_state list) : det_state list =
			List.filter (is_included ds) ld;;
		
		
		(* all combinations of several lists 
     
     all_combinations : 'a list list -> 'a list list 

     example: 
     all_combinations [[1];[2;3];[4]];;
     - : int list list = [[1; 2; 4]; [1; 3; 4]] *)


  let rec all_combinations l=
    let rec add_one_of (l1: 'a list) (l2: 'a list list)=
      match l1 with
	  [] -> []
	| elt::lrem -> (List.map (function x -> elt::x) l2)@(add_one_of lrem l2)
    in
      match l with
	  [] -> [[]]
	| first_comb::lrem -> 
	    let rem_combinations= all_combinations lrem in
	      add_one_of first_comb rem_combinations
		
		(* instanciating a transition by all possible and corresponding deterministic states taken in l. *)
		let instanciate ((f, dsl, ddest): det_symtrans) (l: det_state list) : det_symtrans list =
			if dsl = [] then
				[(f, dsl, ddest)]
			else
				let all_possible_cases = List.map (fun s -> get_containing s l) dsl in
				if List.exists (fun x -> x=[]) all_possible_cases then
					[]
				else
					List.map (fun x -> (f, x, ddest)) (all_combinations all_possible_cases);;

	let rec folder_flatten folder= 
    match folder with
				[] -> []
      | (_,l)::rem -> List.append l (folder_flatten rem);;

  (* union of det_state sets (lists) *)

  let rec list_union l1 l2=
    match l1 with
	[] -> l2
      |s::lrem -> 
	 let union_rem= list_union lrem l2 in
	   if List.mem s union_rem then union_rem else s::union_rem;;

  let rec folder_replace (ltrans: 'a) (criteria: 'b) (folder: ('b * 'a) list)  =
    match folder with
	[] -> [(criteria, ltrans)]
      | (c, trs)::rem when c=criteria -> 
	  (c, ltrans)::rem
      | any::rem -> any::(folder_replace ltrans criteria rem)

(* adding a transition to a deterministic set of transitions (folder). 

     When adding transition (b) after having added transition (a)
     both are merged into f({q1,q2},{q1,q2})->{q3,q4} and this provoques
     the apparition of new det_state {q3,q4}. *)

  let add_det_trans ((f,dsl,ddest): det_symtrans) (folder: (symbol * det_symtrans list) list)  =
    (*      let _= print_string ("Adding "^(det_trans_to_string t)^"\n") in *)
    let rec aux t ltrans new_trans=
      match t,ltrans with
	  (_,_,ds1), [] -> (t::new_trans, [ds1])
	      
	| (sym1, l1, ds1), (_, l2, ds2)::rem when l1=l2 ->
	    if is_included ds1 ds2 then (new_trans@ltrans, [])
	    else 
	      let new_det_state= detstate_union ds1 ds2 in
		(* let _=print_string ("gives: "^(det_trans_to_string (sym1, l1, new_det_state))^"\n\n") in *)
		(new_trans@[(sym1, l1, new_det_state)]@rem, [new_det_state])
		
	| _, t2::rem -> aux t rem (t2::new_trans)
    in 
    let (new_trans, new_det_states) = (aux (f,dsl,ddest) (try List.assoc f folder
					      with Not_found -> []) []) in
      (folder_replace new_trans f folder, new_det_states)
      

  (* adding a list of transition to a deterministic set of transitions *)

    let add_all_det_trans (ltrans: det_symtrans list) (folder: (symbol * det_symtrans list) list)=
      let rec aux ltrans folder new_states=
        match ltrans with
        	  [] -> (folder, new_states)
        		| t::rem -> 
        	    let (new_delta, first_new_states)= add_det_trans t folder in
        	      aux rem new_delta (list_union first_new_states new_states)
      in aux ltrans folder [];;

		let trans_to_det_trans ((f, sl, dest) : external_symtrans) : det_symtrans =
			let dsl = List.map (fun x -> Stateset.singleton x) sl
			and ddest = Stateset.singleton dest
			in (f, dsl, ddest);;

		let determinise (a_orig: automaton) : automaton =
			let a = export (export_epsi a_orig) in
			let resultat = {input = a;
					next_state = 1;
					memo_repr = Hashtbl.create 0;
					memo_repr_rev = Hashtbl.create 0;}
			in
			
			let init_trans = List.map (fun (Trans(t)) -> trans_to_det_trans t) (Transitions.trans_to_list a.trans.transitions) in
			
			let rec aux (ltrans: det_symtrans list) (new_trans: (symbol * det_symtrans list) list) 
      (new_states: det_state list) (added: bool) =
				match ltrans with
					| [] ->
						if not added then
							(folder_flatten new_trans, new_states)
						else
							aux init_trans new_trans new_states false
					
					| one_trans::rem ->
      	    let possible_trans= instanciate one_trans new_states in
        	    let (incremented_new_trans, other_new_states)=
        	      add_all_det_trans possible_trans new_trans in
        	      aux rem incremented_new_trans (list_union other_new_states new_states) 
        					((not (other_new_states = [])) || added)
		
    in let (les_trans, les_etats) = aux init_trans [] [] false
		
		in let les_finaux = List.filter (detstate_is_final resultat) les_etats
		and les_trans_eff = List.map (fun t -> Trans(rp_trans resultat t)) les_trans
		in let les_finaux_eff = Stateset.add_list (List.map (rp_state resultat) les_finaux) Stateset.empty
		in accessibility_cleaning (external_to_TA les_finaux_eff les_trans_eff)
		
end



