open Global
open Term
open Substitution
open Definitions

(* Conditionnal equation system  *)

(* Regular: (f(x)=g(x,_) if x=y 
   Exteq: [f(x,y)=g(_)] -> [g(x)=toto, h(y)=h(x)] 
   Extany: [f(x,y),g(x)] -> [g(x)=toto, h(y)=h(x)] 

   Note that an equation of the form f(x)=g(x) will be encoded by f(x)=g(x) if true and
   [f(x,y)] =>[x=y] will be encoded by [f(x,y)=_] => [x=y]

*)

(* type t = T of int * int different de type t = T of (int * int) *)
type equation =  
  | Regular of (term * term * (term * term) list)
  | Exteq of (term * term * (term * term) list)
  | Extany of (term * term * (term * term) list)

type equational_theory = equation list


exception InclusionTestFailed of string
exception NotHandledCondition of string
exception EquationMember of string
exception ConditionsException of term * term
exception ContextualEquationNotHandledYet

exception ProblemWithConditions of string


(* Variables in conditions of the equation have to appear in the equation.
   hence, no need to look at them *)
let vars : equation -> variable list = function
  | Regular (l, r, _) | Exteq (l, r, _) | Extany (l, r, _) ->
      List2.union (Term.vars l) (Term.vars r)

let free_vars : equation -> variable list = function
  | Regular (l, r, c) | Exteq (l, r, c) | Extany (l, r, c) ->
      let lvars = Term.vars l and rvars = Term.vars r in
      let condvars = 
        List.fold_right 
	  (fun (cl, cr) rem -> 
	     List2.union (Term.vars cr) (List2.union (Term.vars cl) rem)
          )
	  c []
      in
      let lvars2 = List2.diff lvars ("_"::rvars@condvars) in  (* "_" is not free *)
      let rvars2 = List2.diff rvars ("_"::lvars@condvars) in
      let condvars2 = List2.diff condvars ("_"::lvars@rvars) in
	List2.clean (lvars2@rvars2@condvars2)

  
let vars_eq (et: equational_theory) : variable list =
  List.fold_right (fun e -> List2.union (vars e)) et [] 


let conditions_checking conds lvars = 
  List.iter
    (fun (t1, t2) ->
       if 
         not (List.for_all (fun x -> List.mem x lvars) (Term.vars t1 @ (Term.vars t2)))
       then
	 raise (ConditionsException (t1, t2))
    ) conds
  
let equation_to_string (lhs: term) (rhs: term) : string = 
  Term.to_string lhs ^ " = " ^ (Term.to_string rhs)

let to_string : equational_theory -> string = 
  List.fold_left
    (fun str eq -> str ^
       (function
         | Regular (lhs, rhs, cond) -> 
             (equation_to_string lhs rhs) ^ (if cond = [] then "" else " if ") ^
               (List2.to_string ~sep:" /\\ " ~left:"" ~right:"\n"
                  (fun (t1, t2) -> (Term.to_string t1) ^ "!=" ^ (Term.to_string t2))
                  cond
               )
    (* String.concat " /\\ " (List.map (fun (t1, t2) ->
       (Term.to_string t1) ^ "!=" ^ (Term.to_string t2)) cond ) ^
       "\n" ^ (to_string rem) *)
    (* Special treatment of Exteq of the form [s] => ... encoded by an equation
       [s=_] => ... *)
         | Exteq(lhs,Var("_"),cond) ->
             "["^(Term.to_string lhs)^"]"^
               (if cond=[] then failwith "Equation:to_string:invalid Exteq equation"
	        else " => ["^(String.concat ", "
			        (List.map (function (t1,t2)->
			          equation_to_string t1 t2)
			           cond))^"]\n")
               
         | Exteq (lhs, rhs, cond) ->
             "[" ^ (equation_to_string lhs rhs) ^ "]" ^
	       if cond = [] then 
                 failwith "Equation:to_string:invalid Exteq equation"
	       else
                 List2.to_string ~sep:", " ~left:" => [" ~right:"]\n"
                   (fun (t1, t2) -> equation_to_string t1 t2)
                   cond
         | Extany (lhs, rhs, cond) ->
             "[" ^ (Term.to_string lhs) ^ "," ^ (Term.to_string rhs) ^ "]" ^
	       if cond = [] then
                 failwith "Equation:to_string:invalid Exteq equation"
	       else
                  List2.to_string ~sep:" " ~left:" => [" ~right:"]\n"
                    (fun (t1, t2)-> equation_to_string t1 t2)
                    cond
                    
       ) eq
    ) ""

let apply (s: substitution) : equation -> equation = 
  let apply_s (l, r, cond) = 
    let apply_s = Rewriting.subst_apply s in
      (apply_s l, apply_s r, 
       List.map (fun (c1, c2) -> apply_s c1, apply_s c2) cond
      )
  in
    function
      | Regular eq -> Regular (apply_s eq)
      | Exteq eq -> Exteq (apply_s eq)
      | Extany eq -> Extany (apply_s eq)

(* remove unused variables and replace them by anonymous ones *)
let simplify (e: equation) : equation = 
  let freev = free_vars e in
  let renaming =
    List.fold_right (fun v s -> Substitution.add v (Var "_") s)
      freev Substitution.empty 
  in
    apply renaming e
      

(* Check syntax and simplify an equation *)

let check (e: equation) : equation =
  let simpe = simplify e in
    match e with
      | Regular (lhs, rhs, cond) ->
	  if lhs = Var "_" || rhs = Var "_" then 
            (* B: On raconte pas sa vie dans les exceptions!!! Y a le failwith pour ca... *)
	    raise (EquationMember "Left and right hand side of an equation must be different of '_'.")
	  else
            (try
	       conditions_checking cond ((Term.vars lhs) @ (Term.vars rhs) @ ["_"]);
	       (*Term.non_linearity_checking lhs;
	       Term.non_linearity_checking rhs;
					supprimé le 9 janvier 2014 suite à réunion Genet/Salmon *)
	       simpe
	     with 
	       | ConditionsException (t1, t2) -> 
                   raise 
                     (ProblemWithConditions 
                        ("Some variables of the condition " ^ 
                           (Term.to_string t1) ^ " != "^(Term.to_string t2) ^
                           " are not included in the set iof variables extracted from the equation " ^
                           (Term.to_string lhs) ^ " = " ^ (Term.to_string rhs)
                        )
                     )
	       | Term.Non_linear -> 
                   raise
                     (EquationMember
                        ("Left and right hand sides of the equation " ^
                           (Term.to_string lhs) ^ " = " ^ (Term.to_string rhs) ^ " have to be linear. "
                        )
                     )
            )
      | Exteq (Var _,Var _, _) | Extany (Var _, Var _, _) ->
          raise (EquationMember "Both left and right hand side of an equation cannot be reduced to a variable")
      | Exteq (lhs, rhs, tl) | Extany (lhs, rhs, tl) -> 
	  try
	    Term.non_linearity_checking lhs;
	    Term.non_linearity_checking rhs;
	    List.iter
              (fun (t1, t2) ->
                 Term.non_linearity_checking t1;
                 Term.non_linearity_checking t2
              ) tl;
	    if List.exists (fun (t1, t2) -> t1 = Var "_" || t2 = Var "_") tl then
              raise 
                (EquationMember
                   ("Having anonymous or free variables in merging equation lists of contextual equations is forbidden: " ^ 
                      (to_string [e])
                   )
                )
	    else 
              simpe
	  with 
	    | Term.Non_linear -> 
                raise
                  (EquationMember
                     ("Left and right hand sides of the equation "^ (Term.to_string lhs) ^
                        " = " ^ (Term.to_string rhs) ^ " have to be linear. "
                     )
                  )
                
	  

(* Conversion of parser format to equation type of Timbuk *)

let pEquation_to_equation = function 
    (* Regular equations *)
  | PEquation (lhs, rhs, cond) -> 
      let new_lhs = Term.pTerm_to_term lhs
      and new_rhs = Term.pTerm_to_term rhs in 
      let new_conds = 
	List.map
          (function
	     | PNotEqual (t1, t2) -> (Term.pTerm_to_term t1, Term.pTerm_to_term t2)
                 (* B: Encore une exception qui sert a rien !!!!! *)
	     | _ -> raise (NotHandledCondition "Only disequalities are expected in conditions of equations.")
	  ) 
	  cond 
      in 
	check (Regular (new_lhs, new_rhs, new_conds))
          (* Contextual or Extended equations *)
  | PExteq (lhs, rhs, tl) ->
      let new_lhs = Term.pTerm_to_term lhs
      and new_rhs = Term.pTerm_to_term rhs in 
      let new_tl =
        List.map (fun (t1, t2) -> Term.pTerm_to_term t1, Term.pTerm_to_term t2) tl
      in
	if tl = [] then
          (* B: Encore une exception a virer !!!!*)
          raise (EquationMember ("List of merging equations of contextual equation ["^(Term.to_string new_lhs)^"="^(Term.to_string new_rhs)^"] -> [] cannot be empty!"))
	else 
	  check (Exteq (new_lhs, new_rhs, new_tl))

  | PUnique (lhs, tl) ->
      let new_lhs = Term.pTerm_to_term lhs in 
      let new_tl = List.map (fun (t1, t2) -> Term.pTerm_to_term t1, Term.pTerm_to_term t2) tl in
	if tl = [] then
          raise (EquationMember ("List of merging equations of contextual equation ["^(Term.to_string new_lhs)^"] -> [] cannot be empty!"))
	else 
	  check (Exteq (new_lhs, Var "_", new_tl))
  | PAny (lhs, rhs, tl) ->
      let new_lhs = Term.pTerm_to_term lhs in 
      let new_rhs = Term.pTerm_to_term rhs in 
      let new_tl= List.map (fun (t1, t2) -> Term.pTerm_to_term t1, Term.pTerm_to_term t2) tl in
	if tl = [] then
          raise (EquationMember ("List of merging equations of contextual equation [" ^
                                   (Term.to_string new_lhs) ^ "=" ^(Term.to_string new_rhs) ^
                                   "] -> [] cannot be empty!")
                )
	else
	  check (Extany(new_lhs, new_rhs, new_tl))



