(** Dico est un ensemble fini de mots *)
module type Language = sig
  type char
  val compare: char -> char -> int

  type word
      
  val hd : word -> char
  val tl : word -> word
  val empty : word

  val cons : char -> word -> word
  val is_empty : word -> bool
end

(** Example of language *)
module Lang: Language = struct
  type char = int
  let compare = compare
    
  type word = char list
  let hd = List.hd
  let tl = List.tl
  let empty = []
  let cons = (fun x y -> x::y)
  let is_empty = (=) []
end

module Make (Lang: Language) = struct
  
  type word = Lang.word
      
  type state = int

  module CMap = Map.Make (struct type t = Lang.char let compare = Lang.compare end)
  module SMap = Map.Make (struct type t = state let compare = compare end)


  type t = 
      { mutable card    : state;
        mutable finals  : unit SMap.t;
        next    : state CMap.t SMap.t; 
        prev    : state CMap
      }

  let init_st = 0

  let empty = { card = 0; finals = SMap.empty; next = SMap.empty; prev = SMap.empty }
    (** The empty dictionnary. *)
    
  let is_empty auto = auto.card = 0 || SMap.is_empty auto.finals
    (** Test whether a map is empty or not. *)

  let is_final auto st =
    SMap.mem st auto.finals

  let set_final auto st =
    SMap.add st () auto.finals

  let unset_final auto st
      SMap.remove st auto.finals
 
  let fresh_state auto =
    let st = auto.card in
      auto.card <- auto.card + 1; st

  let next auto st ch =
    CMap.find ch (SMap.find st auto.next)

  let prev auto st ch =
    CMap.find ch (SMap.find st auto.prev)


  (** *Only used for benchmarks
     Returns [true] if all states of an automaton [auto] are defined
     and if all transition set and its mirror are consistents.
     For an automaton [auto],
        - the state [q] is defined iff [q < auto.card];
        - transition sets [s1] and [s2] are consitent iff for all states
          [q], [q'] and character [c], transition [(q, c, q')] is in [s1]
          is equivalent to transition [(q', c, q)] is in [s2].
  *)
  let check_auto auto =
    let i = ref 0 in
    let step1 = (* check definition of states and the "==>" side of consitence *) 
      SMap.forall
        (fun q trans -> 
           CMap.forall
             (fun c q' ->
                incr i;
                q < auto.card && q' < auto.card && (* q and q' are defined *)
                  (try 
                     prev auto q' c  = q
                   with Not_found -> false) (* auto.trans \subseteq auto.prev *)
             )
        )
    in 
      step1 &&
        !i = (SMap.fold (fun _ trans -> (+) (CMap.cardinal trans)) auto.prec 0)
        (* |auto.trans| = |auto.prec| =>
             auto.trans \subseteq mirror(auto.prev) => consistence
        *)


  (** Returns a new automaton such that [auto.prec], the mirror of
      [auto.next], is rebuild. *)
  let mirror auto =
    { auto with prec = 
        SMap.fold
          (fun q ->
             SMap.fold 
               (fun c q' ->
                  SMap.add q' 
                    (CMap.add q' c (try 
                                      SMap.find q' mirror
                                    with Not_found -> CMap.empty
                                   )
                    )
               )
          )
          auto.prec
          SMap.empty
    }


  let rec rm_state auto st =
    try (
      if (st < auto.card) then
        (* We rename [st'], the last state added to [auto], in [st] the state to remove. *)
        let st' = auto.card - 1 in
          try (
            let auto = 
              if is_final auto st' auto then
                unset_final (set_final auto st) st'
              else
                unset auto st
 (SMap.find st' auto.final) auto.final 
              with Not_found -> SMap.remove 
            let next = SMap.add st (SMap.find st' auto.next) auto.next
            let prec = SMap.add st (SMap.find st' auto.prec) auto.prec
          with Not_found -> 
            auto [TODO Mettre a jour  ] 
              (* [st'] n'a pas de transition... attention il faut quand meme mettre a jour*)
          )
      else
        (* Here, we try to remove a state which is not defined i.e.  
           [ not (st < auto.card) ]. Thus, [auto] is exactly returned *)
        auto
        (* assert false or raise (Not_a_state st) *)
        
      let st' = auto.card - 1 in
      let trans 
          
    ) with
      | Not_found -> 
          auto.card <- auto.card - 1;
          rm_fwd_trans auto 

    
  let clean auto = 
    let clean trans =
          


    and bwd_clean trans =


    in
      bwd_clean fwd_clean auto.trans


  let rec mem auto st wd =
    if (Lang.is_empty wd) then
      is_final auto st
    else
      try 
        let st' = next auto (hd wd) in
          (* raise Not_found if there is no transition of the form st --ch--> st' *)
          mem st' (tl wd)
      with Not_found -> false
        (** [mem d w] returns [true] if [d] contains the word [w]
            and [false] otherwise. *)

  let singleton wd = 
    if (Lang.is_empty wd) then
      { empty with final = true }
    else
      { trans = CMap.add (hd wd) { new_st (tl wd) with prefix = wd } CMap.empty;
        final = false
      }
        (** [singleton w] returns a fresh state such that [mem (singleton w) w]
            returns [true] and [false] for any other word different from
            [w]. *)
          
  let add auto wd =
    let rec add_aux st wd =
      if (Lang.is_empty wd) then
        { auto with final = true }
          (* if [st] is a final state then the emptyWord is accepted *)
      else
        try 
          let st' = CMap.find (hd wd) st.trans in
            (* Not_found is raised, only if there is no transition of the form st
               --(hd wd)--> st' *)
            { st with
                trans = CMap.add (hd wd) (add st' (tl wd)) st.trans 
          }
        with Not_found -> (* Here, we know that wd is not empty *)
          { st with 
              trans = CMap.add (hd wd) (new_st wd) st.trans 
          }
    in
      add_aux auto.init wd
            (** [add d w] returns a dictionnary containing the same words as
                [d] plus the word [w]. *)
            
            
  let rec remove st wd =
    if (Lanf.is_empty wd) then
      { st with final = false }
        (* [st] does not accept the [emptyWord], only if [final] is
           set to [false] *)
    else
      try 
        let st' = CMap.find (hd wd) st.trans in
          (* Not_found is raised, only if there is no transition of
             the form st --(hd wd)--> st' *)
        let st'' = remove st' (tl wd) in
          if is_empty st'' then
            (* if [st''] does not contain anything, we forget it, else
               we keep it *)
            { st with trans = CMap.remove (hd wd) st.trans } 
          else
            { st with trans = CMap.add (hd wd) st'' st.trans } 
      with Not_found -> st
        (* [wd] is not in [st]: we do nothing *)
        (** [remove d w] returns a dictionnary that contains all words contained in [d],
            except of [w] which is no longer in the returned dictionnary. *)
        
  let rec compare st st' =
    if (compare st.final st'.final <> 0) then
      compare st.final st'.final
    else
      CMap.compare compare st.trans st'.trans
        
  let rec equal st st' =
    (st.final = st'.final) &&
      CMap.equal equal st.trans st'.trans

  let iter f st =
    let rec iter_aux wd st =
      if st.final then f wd;
      CMap.iter (fun ch st' -> iter_aux (Lang.cons ch wd) st') st.trans
    in
      iter_aux Lang.empty st
        
  let fold f st init =
    let rec fold_aux wd cur st
        if st.final then
          f wd cur
        else
          CMap.fold
            (fun ch st' cur' -> 
               fold_aux (Lang.cons ch wd) cur' st' 
            )
            st.trans
            cur
    in
      fold_aux Lang.empty init st
            
  let for_all p st = 
    let rec for_all_aux wd st =
      (if st.final then p wd else true) &&
        CMap.for_all (fun ch st' -> for_all_aux (Lang.cons ch wd) st')
    in
      for_all_aux Lang.empty st
        
  let exists p st =
    let rec exists_aux wd st =
      (if st.final then p wd else false) ||
        CMap.exists (fun ch st' -> exists_aux (Lang.cons ch wd) st')
    in
      exists_aux Lang.empty st

  let filter p st =
    let rec filter_aux wd st =
      { final = st.final && (p wd);
        trans =
          CMap.fold
            (fun ch st cmap ->
               let st' = filter_aux (Lang.cons ch wd) st in
                 if (is_empty st') then
                   cmap
                 else
                   CMap.add ch 
             
                   
        

end
