open Dumper

(** This module is an extension of module Hashtbl *)
include Hashtbl

let length (tbl: ('a, 'b) t) : int =
  Hashtbl.fold (fun _ _ k -> k + 1) tbl 0

(* Former hashtbl_to_list *)
(** Returns an association list [elements t] such that for all pair
    [(key, data)], [t] maps [key] to [data] is equivalent to the pair
    [(key, data)] belongs to [elements t].
*)
let elements (t: ('a, 'b) Hashtbl.t) : ('a * 'b) list =
  Hashtbl.fold (fun i j l -> (i, j) :: l) t []

(** [map f tbl] returns a new hashtbl that contains
    [(a, f b)] iff [tbl] contains the pair [(a, b)].
    The input hashtable [tbl] is not modified.
*)
let map f tbl =
  let tbl' = create (length tbl) in
    iter (fun a b -> add tbl' a (f a b)) tbl;
    tbl'

(** [umap f tbl] same as [map f tbl] but in imperative style. It replaces
    each pair [(a, b)] of [tbl] by [(a, f a b)].
*)
let umap f t =
  iter (fun a b -> replace t a (f a b)) t



(** Parametrizable formater to display hash tables. To use the default
    display just invoke [to_string] using the appropiate formater
    corresponding to for keys and stored values. For example,
    [to_string string_of_int (fun x -> x)] can be applied to any hash
    table of type [(int, string) Hashtbl.t]. It returns a string of
    the form ["{(k1, v1); ... ; (kn, vn)}"]. As for {!List2.to_string},
    semi-colon, left and right marker can be specified by using
    optional parameters syntax:
    [to_string ~sep:" | " ~left:"[[ " ~right:" ]]"]
    All is definitely parametrizable by specifying a particular formater
    for each pair of (key, value):
    [ let my_format str_key str_val key value =
        "{" ^ (str_key key) ^ " -> " ^ (str_val value) ^ "}"]
    Thus, [to_string ~sep:" | " ~left:"[[ " ~right:" ]]" ~format:format] returns
    strings of the form ["[[ {k1 -> v1} | ... | {kn -> vn} ]]"]
*)
let to_string ?(sep = "; ") ?(left = "{") ?(right = "}") ?(format) =
  fun ?(str_key = Dumper.dump) ?(str_val = Dumper.dump) htbl ->
    let format =
      match format with
        | None ->
            (fun key data ->
               "(" ^ (str_key key) ^ ", " ^ (str_val data) ^ ")")
        | Some f -> f str_key str_val
    in
    let str =
      Hashtbl.fold (fun k d str -> str ^ (format k d) ^ sep) htbl left
    in
      (if str = left then
         left
       else (* str = "{(k1, d1); ... ; (kn, dn);"
               the last [sep] must be removed  ^ *)
         String.sub str 0 (String.length str - (String.length ) sep)
      ) ^ right



(** Exception [Break] is used to simulate break instruction as in
    imperative style *)
exception Break

(** [subset equal t1 t2] returns [true] iff [t1] is a subset of [t2] *)
let subset equal t1 t2 =
  try
    Hashtbl.iter
      (fun a b -> if not (equal (Hashtbl.find t2 a) b) then raise Break)
      t1;
    true
  with
    | Not_found | Break -> false


let equal equal t1 t2 : bool =
  subset equal t1 t2 && subset equal t2 t1


(** [no_double tbl] returns [true] if and only if the hash table [tbl] does
    not contains any double
*) (* former [check t] *)
let no_double tbl = List2.no_double (elements tbl)



IFDEF TEST THEN
open Debug
ENDIF
(* Copy a Hashtbl mapping keys to Hashtbl *)
let copy_level2 tbl =
(* hashtbl_copy works only if there is only one occurence of each key in
   the table *)
  let _ =
    IFDEF TEST THEN
      asserteq (no_double tbl) true "No double keys in hashtbl we use"
    ELSE
      ()
    END
  in
    map (fun _ -> copy) tbl


(* Copy a Hashtbl mapping keys to Hashtbl mapping keys to Hashtbl *)
let copy_level3 tbl =
(* hashtbl_copy works only if there is only one occurence of each key in
   the table *)
  let _ =
    IFDEF TEST THEN
      asserteq (no_double tbl) true "No double keys in hashtbl we use"
    ELSE
      ()
    END
  in
    map (fun _ -> copy_level2) tbl



(* esc is an escaping string such as "   " in order to indent if necessary (useful for
recursive calls with to_string_level2,3 etc *)
let to_string2 esc to_str tbl =
  fold
    (fun k e str -> str ^ esc ^ (Dumper.dump k) ^ " -> " ^ (to_str e) ^ "\n")
    tbl ""

let to_string_level1 esc tbl =
  to_string2 esc Dumper.dump tbl

let to_string_level2 esc tbl =
  fold
    (fun k t str -> str ^ esc ^ (Dumper.dump k) ^ " -> \n" ^
       (to_string_level1 (esc ^ "      ") t) ^ "\n")
    tbl ""

let to_string_level3 esc tbl =
  fold
    (fun k tt str -> str ^ esc ^ (Dumper.dump k) ^ " -> \n" ^
       (to_string_level2 (esc ^ "      ") tt) ^ "\n")
    tbl ""

let equal_level1 eq t1 t2 =
  equal eq t1 t2

let equal_level2 eq t1 t2 =
  equal (equal eq) t1 t2

let equal_level3 eq t1 t2 =
  equal (equal (equal eq)) t1 t2
