(* ocamlbuild -cflags -dtypes -pp "camlp4o pa_macro.cmo -DTEST" main.byte *)
(* ocamlbuild -pp "camlp4o pa_macro.cmo -UTEST" main.native *)
open Term
open Equation
open Automaton

(* Property type: pattern or Automaton (automaton permits the automatic refinement) *)
type property = Pat | Aut of automaton

let delete_file (filename: string): unit=
  try (Sys.remove filename)
  with (Sys_error _) -> ()

let write_file
    (filename: string) (alpha: (string*int) list) (varset: string list)
    (trs: Rewriting.trs) (ainit: Automaton.automaton) (pl: term list)
    (el: equational_theory) (acomp: Automaton.automaton) : unit =
  let resstring =
    "% Completion time: "^ (string_of_float (Sys.time() -. !Common.start_time)) ^
      " seconds\n\n" ^ "Ops " ^
      (List.fold_right
         (fun (s, i) rem -> s ^ ":" ^ (string_of_int i) ^ " " ^ rem)
         alpha
         ("\n\nVars "^
	    (List.fold_right
               (fun x rem -> x ^ " " ^ rem)
	       varset
	       ("\n\nTRS R1\n" ^ (Rewriting.to_string trs) ^
	          "\n\nAutomaton Ainit\n" ^ (Automaton.to_timbuk ainit) ^
	          "\nAutomaton Acomp\n" ^ (Automaton.to_timbuk acomp) ^
	          (if List.length pl = 0 then
                     ""
                   else if !Common.subpattern then
                     "\nSubPatterns\n"
                   else
                     "\nPatterns\n"
                  ) ^ (Term.list_to_string pl) ^ "\n"
               ) ^
	       if List.length el = 0 then "" else
	         "Equations Approx\nRules\n" ^ (Equation.to_string el)^"\n"
            )
         )
      )
  in
  let outchan = open_out filename in
    try
      Printf.fprintf outchan "%s" resstring;
      close_out outchan
    with exc -> close_out outchan; raise exc



(* Check only the intersection, no completion step *)
let check_only ta pl trs_and_pl_vars al =
  match !Common.property_type with
    | Common.Patterns ->
        let allres = Completion.check_intersection ta pl trs_and_pl_vars in
          (* If not SubPatterns are considered we focus on final states and
	     predecessors w.r.t. epsilon trans of final states *)
        let allfiltered =
          if !Common.subpattern then allres else
            Completion.filter_final
              allres
              (Transitions.get_all_predPhi ta.fstates ta.trans)
              (Transitions.get_all_pred ta.fstates ta.trans)
        in
          (* no completion, only pattern check *)
	  print_string ("\n\nPattern check:\n--------------\n\n" ^
		          (Completion.patternres_to_string allfiltered) ^ "\n");
	  (false, ta, 0, [], false, Automaton.make ()) (*Formula.mk_false*)
             (* don't care of the result in that case *)

    | Common.Automata propname ->
        let prop_ta =
          try Dataextracting.autassoc propname al with
            |  Dataextracting.Automaton_not_found ->
                 failwith ("Main:first_glance: No automaton of this name "^propname)
        in
        let phi = Automaton.inter_of' ta prop_ta in
	  if not(Formula.is_sat phi) then (
	    print_string "empty intersection = no forbidden term is reachable\n";
            (false, ta, 0, [], false (* Formula.mk_false*), Automaton.make ()) (* don't care of the result in that case *)
	  ) else (
	    print_string "At least one forbidden term is reachable\n";
	    (false, ta, 0, [], (* Formula.mk_false*) false, Automaton.make ())
          )


let write_step alpha varset trs ta pl equ dpl (outfile_basename: string) (num_step: int) (intermediaire: bool) (the_step: Completion.step) : unit = 
	let fnameres= if intermediaire then outfile_basename^(string_of_int num_step) else outfile_basename in
    Verbose.exec 2 
      (function _ -> 
         print_string ("Done\nExport "^fnameres^".full file...");
         (write_file (fnameres^".full") alpha varset trs ta pl equ the_step.Completion.aredex));
    print_string ("Done\nExport "^fnameres^".res.epsi file...");
    (try 
       (write_file (fnameres^".res.epsi") alpha varset trs ta pl equ 
          (snd(Automaton.renumerotation (Automaton.utility_cleaning(Automaton.accessibility_cleaning(Automaton.export_epsi the_step.Completion.aredex))))))
     with e -> print_string "\nCannot write .res.epsi file!\n";
       raise e);
    print_string ("Done\n");
    if !Common.export then
      begin
        print_string ("Export "^fnameres^".res file...");
        (try 
           (write_file (fnameres^".res") alpha varset trs ta pl equ
              (Automaton.export the_step.Completion.aredex))
         with e -> print_string "\nCannot write .res file!\n";
           raise e);
        print_string "Done\n"
      end
    else
      delete_file (fnameres^".res");
    print_string ("Export "^fnameres^".dot file...");
    (try
       Dot.write_dot (fnameres^".dot") the_step.Completion.aredex dpl
     with e -> print_string "\nCannot write .dot file!\n";
       raise e);
    print_string "Done\n"

(* Main function call for completion: filename and maximal number of
   completion steps. Results into a bool saying if the automaton is
   complete, the completed automaton, the number of the last performed
   completion step, a list of found patterns and a bool saying if a counterexample has been found
*)

let first_glance (filename: string) (outfile_basename:string)(autname: string) (propname: string) (nbstep: int):
    (bool * Automaton.automaton * int * ((Term.term * Definitions.state * Substitution.substitutionset) list) * bool * Automaton.automaton) =

  let (trsl, al, pl, dpl, el) = Dataextracting.extraction filename in
  let alpha = !Global.alpha and varset = !Global.varset in
    match trsl, al with
      | [], _ -> failwith "Main.first_glance: Missing Term Rewriting System\n"
      | _, [] -> failwith "Main.first_glance: Missing Tree Automaton\n"
      | (_, trs)::_, (_, tmpta, _)::_ ->
	  let ta = (* ta est l'automate A0 ; il est repéré par autname dans le fichier ; avec un automate par défaut si cette chaine est vide *)
            if autname = "" then tmpta
            else
              try Dataextracting.autassoc autname al with
	        | Dataextracting.Automaton_not_found ->
                    failwith "Main.first_glance: No automaton of this name"
	  in
		let airr = try Dataextracting.autassoc "AIRR" al with Dataextracting.Automaton_not_found -> failwith "Main.first_glance: AIRR not in spec" in
	  let equ = try snd (List.hd el) with _ -> [] in
	    (* get all variables in patterns + trs on one side and equations *)
	  let trs_and_pl_vars =
            List.fold_right (fun t -> List2.union (Term.vars t))
	      pl (Rewriting.vars_trs trs)
          in
	  let eqvars = Equation.vars_eq equ in
	    if !Common.check_only then
              check_only ta pl trs_and_pl_vars al
	    else (* full completion *)
              let c = (* B: OK, l'algo commence seulement ici *)
                ref (Completion.new_step ta trs equ pl trs_and_pl_vars eqvars airr) (* c'est cette fonction qui s'occupe de créer Alhs ! d'où son nom, sans doute *)
              in
              let prop =
                if propname = "" then
                  Pat
                else
                  try Aut (Dataextracting.autassoc propname al) with
	            | Dataextracting.Automaton_not_found ->
                        failwith ("Main.first_glance: No automaton of this name " ^ propname )
              in
	      let i = ref 1 and pat = ref [] and failed = ref false in
                (* and inter = ref Formula.mk_false in *)
	      let olddelta = ref (Transitions.make ()) in
	        (* begin *)
                Verbose.print 3 (fun _ -> "init cs=" ^ (Completion.to_string !c) ^"\n\n");
		while
                  (* tant que des transitions sont produites *)
                  not (!c.Completion.new_trans = [] && !c.Completion.new_epsi = []) &&
                    (* Tant qu'il n'y a pas de contre-exmples *)
                    not !failed && (* (Formula.tautology !inter) && *)
                    (* Tant que la limite n'est pas depassee *)
                    !i <= nbstep &&
                    (* tant qu'aucun pattern interdit n'a pas ete trouve... *)
                    !pat = []
                do
                  Verbose.println 3 (fun _ -> "Before computation: " ^ (Automaton.to_string !c.Completion.aredex));
                  let t1 = Sys.time () in
		  let (c2, pat2) = Completion.do_one_step !c in
		  let t2 = Sys.time () in
                    Verbose.println 3 (fun _ -> "Right after computation: " ^ (Automaton.to_string c2.Completion.aredex));
		    print_string ("\nCompletion step done in: "^(string_of_float (t2 -. t1))^" seconds\n");
                    print_string ("\n\nCompletion step: "^(string_of_int !i)^
				    "\n---------------------------------------------------------\n");
                    Verbose.print 3 (fun _ -> "Completion step at: " ^ (string_of_int !i) ^
				       ":\n" ^ (Completion.to_string c2) ^ "\n\n");
                    Verbose.println 2 (fun _ -> "Automaton at step: " ^ (string_of_int !i) ^
                                         "\n" ^ (Automaton.to_string c2.Completion.aredex));
		    print_string ("New transitions: "^(string_of_int (List.length c2.Completion.new_trans))^"\n");
		    print_string ("New epsilon: "^(string_of_int (List.length c2.Completion.new_epsi))^"\n\n");
		    flush stdout;
		    Verbose.print 0
                      (fun _ -> "Transitions to add: " ^
                         (Transitions.symextlist_to_string c2.Completion.new_trans)^"\n\n");
		    Verbose.println 0
                      (fun _ -> Transitions.epsiextlist_to_string c2.Completion.new_epsi);
		    pat := pat2;
                    Verbose.print 0
                      (fun _ -> Transitions.to_string c2.Completion.aredex.trans);
                    if !Common.interactive then ignore(Unix.system "read");
                    (*  0 => aredex raffiné;
                     *  1 => pas de collision avec a_prop;
                     * -1 => il existe un contre-example *)
                    let ret =  (* if not(!i mod 6 = 0) then 1 else *)
                      match prop with
                        | Pat -> 1 (* AH si y a pas d'automate y a pas de raffinement...
                                           donc on suppose que l'intersection est vide *)
                        | Aut a_prop ->
                            Verbose.println 3
                              (fun _ -> "Automaton aredex for intersection : "^
                                 (Automaton.to_string c2.Completion.aredex));
                            Verbose.println 3
                              (fun _ -> "Automaton prop for intersection : "^
                                 (Automaton.to_string a_prop));
                            let t1 = Sys.time () in
                            let ret = Automaton.refine c2.Completion.aredex a_prop in
                            let t2 = Sys.time () in
                              if ret = 0 then
                                Format.printf "Refinement done in %f\n@?" (t2 -. t1)
                              else
                                Format.printf "Intersection done in %f\n@?" (t2 -. t1);
                              ret
                    in
                      if !Common.interactive then ignore (Unix.system "read");
                      failed := ret = -1;
                      if ret = -1 || ret = 1 then ( (* si on n'a pas raffiné *)
                        (* if not (Formula.is_sat !inter) || (Formula.tautology !inter) then ( *)
                        olddelta := !c.Completion.aredex.Automaton.trans;
                        c := c2;
												if !Common.opt_out_each_step then
													write_step alpha varset trs ta pl equ dpl outfile_basename !i true !c;
                        incr i
                          (* B: counter [i] is only incremented if verification does not fail!!! *)
                      ) else (
                        c := Completion.reset_step c2;
                        Verbose.exec 3 (
                          fun _ ->
                            print_string ("c after reset= "^(Completion.to_string !c)^"\n");
                            print_string ("aredex after reset "^(Automaton.to_string !c.Completion.aredex)^"\n")
                        )
                      )
                done;
                if !failed then
                  Format.printf "Found a counterexample!\nIntersection with automaton %s not empty!\n@?" propname;
                if not !Common.opt_out_each_step then write_step alpha varset trs ta pl equ dpl outfile_basename !i false !c;								
                  ((!c.Completion.new_trans = [] && !c.Completion.new_epsi = []),
                   !c.Completion.aredex, !i - 1,  !pat, !failed, airr)
                    (* end *)

                    (* Main program *)
                    (* Main program can be found after unit testing for clarity reasons (outputs of
                       unit testing must appear before normal outputs of the program *)




                    (* for debug with -g

                       first_glance "nspk2.txt" 5
                    *)




(* ------------------------------ Main program ------------------------------ *)

let parselist (s: string) : int list =
	if Str.string_match (Str.regexp "\\[\\(.*\\)\\]") s 0 then
		let ss = Str.matched_group 1 s in
		let ls = Str.split (Str.regexp ",") ss
		in List.map int_of_string ls
	else
		let max = int_of_string s in
		List2.sequence 1 max;;


exception Unknown_option of string

let rec apply_option_list : string list -> unit = function
  | [] -> ()
  | "--new"::rem -> Common.completion_strategy:=Common.New;
      apply_option_list rem
  | "--prior"::rem -> Common.completion_strategy:=Common.Prior;
      apply_option_list rem
  | "--epsi"::rem -> Common.completion_type:=Common.Epsi;
      apply_option_list rem
  | "--noepsi"::rem -> Common.completion_type:=Common.Noepsi;
      (* In noepsi mode we also merge loops *)
      Common.merge_loop:=true;
      apply_option_list rem
  | "--interactive"::rem -> Common.interactive:=true;
      apply_option_list rem
  | "--equation"::rem -> Common.approx_type:=Common.Equationnal;
      apply_option_list rem
  | "--noapprox"::rem -> Common.approx_type:=Common.None;
      apply_option_list rem
  | "--check"::rem -> Common.check_only:= true;
      apply_option_list rem
  | "--export"::rem -> Common.export:= true;
      apply_option_list rem
	| "--export-fn"::rem -> Common.inter_nf := true;
			apply_option_list rem
  | "--autname"::s::rem -> Common.autname:=s;
      apply_option_list rem
  | "--propname"::s::rem -> Common.property_type:=Common.Automata s;
      apply_option_list rem
	| "--outfile"::s::rem -> Common.opt_outfile_basename := s;
			apply_option_list rem
	| "--outeach"::rem -> Common.opt_out_each_step := true;
			apply_option_list rem

  | "--verbose"::s::rem ->
      (try
				Common.verbose := parselist s;
       with Failure _ ->
         failwith ("--verbose option expects an int or a square-bracket-enclosed list of ints between 0 and "^ (string_of_int Common.max_verbose)^" as parameter, given: "^s)
      );
			apply_option_list rem
  | "--dotson"::rem -> Common.dotson:=true;
      apply_option_list rem
  | "--nodotson"::rem -> Common.dotson:=false;
      apply_option_list rem
  | "--exact"::rem -> Common.exact_completion:=true;
      apply_option_list rem
  | ["--autname"] -> failwith "--autname option should be followed by an automaton name"
  | ["--propname"] -> failwith "--propname option should be followed by an automaton name"
  | ["--verbose"] -> failwith ("--verbose option should be followed by an int or a square-bracket-enclosed list of ints between 0 and "^ (string_of_int Common.max_verbose))
  | ["--outfile"] -> failwith "--outfile option should be followed by a valid file base name"
	| s::_ -> raise (Unknown_option s)

let online_help =
"Usage= timbuk [option] <max number of completion steps> <timbuk spec file>
with options=
             --exact                        : uses exact normalization strategy
             --noapprox                     : does not use approximations defined
                                              in the file spec
             --noepsi                       : uses old style completion and also
                                              sets --mergeloop to true
             --autname <name in spec file>  : name of automaton to use as initial
             --propname <name in spec file> : name of automaton to use as property
             --outfile <basename>           : base name of generated files
             --outeach                      : output each completion step
             --export                       : triggers exportation without epsilon 
                                              transitions
             --export-fn                    : exports Acomp ∩	Airr																 
             --check                        : does not perform completion but
                                              only search for patterns in initial
                                              tree automaton
             --new                          : uses only new states to normalize
             --verbose [<v1>,<v2>,…]        : activates verbose modes v1, v2, … (between 1 and "^(string_of_int Common.max_verbose)^")
             --verbose <v>                  : same as --verbose [1,2,3,…,v]
             --verbose 0, --verbose []      : (default) no verbosity
             --dotson                       : direct subterm edges are exported
                                              the comp.dot file in red

             --epsi                         : (default) uses new style completion with
                                              epsilon transitions
             --equation                     : (default) uses approximation equations
             --prior                        : (default) uses existing states/transitions
                                              when possible
             --nodotson                     : (default) subterm edges are not exported
"

let arg_number = Array.length Sys.argv in
  if arg_number = 1 then (
    print_string online_help;
    exit 0
  );
  try (let l = ref [] in
    (* Ca ne serait pas plus precis de commencer les tests apres le parsing et l'extraction ??? *)
    (* Common.start_time := Sys.time(); *)
    (* C'est quoi l'interet de convertir un tableau en liste ???? *)
    if arg_number > 2 then (
      for i = arg_number - 3 downto 1 do
        l:= Sys.argv.(i)::!l
      done;
      apply_option_list !l
    );

    let nb =
      try int_of_string Sys.argv.(arg_number - 2) with
        | Failure "int_of_string" ->
            print_string ("Bad number of completion steps: " ^
                            (Sys.argv.(arg_number - 2))^"\n");
            exit 0 (* Valeur retournee dans la version precedente *)
		in
    let file = Sys.argv.(arg_number - 1) in
      if not (Sys.file_exists file) then (
        print_string "File does not exist!\n";
        exit 0
      );
      let propname = match !Common.property_type with
      	| Common.Patterns -> ""
      	| Common.Automata yoh -> yoh (* :-) *)
      in
	let outfile_basename = if !Common.opt_outfile_basename <> "" then
														!Common.opt_outfile_basename
													else
														Filename.chop_extension file
	in
	try (
	  let (completed, auto, i, patres, phi, airr) =
            first_glance file outfile_basename !Common.autname propname nb
	  in
	    if not !Common.check_only then (
              (* On ne peut pas afficher l'automate quand les formules sont trop grosses *)
	      (*	  print_string (Automaton.to_string auto); (* Automaton.unepsilon a; *) *)
	      print_string ("\nStep: "^(string_of_int i)^"\n");
	      if completed then
		print_string "\n\nAutomaton complete!\n------------------\n";
	      if patres <> [] then (
		print_string "\n\nPattern found!\n--------------\n";
		print_string (Completion.patternres_to_string patres);
		if !Common.property_type= Common.Patterns then print_string "NO REFINEMENT!\nFor automatic refinement use the --propname option\n"
	      ) else if (*Formula.tautology phi*) phi then
		begin
		  print_string ("\nFormula:"^(string_of_bool phi)^"\n");
		  print_string "\nCounterexample found!\n---------------------\n\n";
		  if !Common.property_type= Common.Patterns then print_string "No property automaton was given, automatic refinement was not possible!\n"
		end
              else
		begin
		  if completed
		  then print_string "\nProof done!\n-----------\n\n";
		  print_string ("\nFormula:"^(string_of_bool phi)^"\n");
		end;
              let fnameres= outfile_basename in
		print_string "\nExported ";
                if !Common.export then print_string (fnameres^".res, ");
                print_string (fnameres^".full and "^fnameres^".dot files\n");
		print_string ("Completion time: " ^ (string_of_float (Sys.time() -. !Common.start_time))^" seconds\n");
		print_string ("Time in Rclosure computations: " ^ (string_of_float !Common.rclosure_time)^" seconds\n");
		print_string ("Time in approximation closure computations: " ^ (string_of_float !Common.approx_time)^" seconds\n");
	    
		if !Common.inter_nf then (
			let anf = Airr.inter_nf auto airr in
			let resstring = Automaton.to_timbuk anf in
			
			let outchan = open_out (outfile_basename^".nf") in
    try
      Printf.fprintf outchan "%s" resstring;
      close_out outchan
    with exc -> close_out outchan; raise exc
			
		)
			
			)
	) 
	with 
	  | Dataextracting.SpecialStateAsRHS -> 
	      begin 
		print_string "\nThe state qTF must not appear in the RHS of a transition\n";
		exit(-1);
	      end
	  | Dataextracting.SpecialStateInEpsilon -> 
	      begin 
		print_string "\nThe state qTF must not appear in a rewriting transition\n";
		exit(-1);
	      end
	      
	  | Dataextracting.SpecialStateDeclaration -> 
	      begin 
		print_string "\nThe state qTF is a special state that can not be declared\n";
		exit(-1);
	      end
	)
	with Unknown_option s -> print_string ("Unknown Timbuk option : "^s^"\n"); print_string online_help; exit 0
