(* A unique error number to each error *)
let errorNumber = ref 0

(* Error style Output "à la Emacs" or "à la Boichut" *)
(* Boichut's style may correspond to an existing style, but I have
   never seen it...  Yohan, you should precise it :-)

   Advantage of Emacs style: when timbuk is executed using
   compilation mode in Emacs, the stderr of the command is parsed and
   errors are automatically located in the corresponding file!
*)
let emacsStyle = ref true

open Lexing
(* Return the position of the current token as a 4-uplet of the form
   file, line and columns of beginning and ending
*)
let localize lexbuf = 
  let p1 = lexbuf.lex_start_p
  and p2 = lexbuf.lex_curr_p in
    (p1.pos_fname, p1.pos_lnum,
     p1.pos_cnum - p1.pos_bol,
     p2.pos_cnum - p2.pos_bol
    )  

(* Handling system errors. *)
(* Mainly when the input specification file cannot be read *)
let handleSystemError errorMessage =
  prerr_endline (errorMessage);
  exit (-1)

(* Handling lexer errors. *)
(* It generates appropriate error messages containing the line number *)
(* where the error has occurred *)
let handleLexerError lexbuf x errorMessage =
  let f, l, c, d = localize lexbuf in
    if !emacsStyle then (
      Format.eprintf "File \"%s\", line %i, characters %i-%i:\n" f l c d;
      Format.eprintf "Bad character : \"%s\" \n@?" (Char.escaped x);
    ) else (
      prerr_string "Line "; prerr_int l;
      prerr_string ": Lex.Err: ";
      prerr_string errorMessage; prerr_string " ";
      prerr_endline (Char.escaped x);
    ); 
    exit (-1)



(* Handling parser errors. *)
(* It generates error messages according to the reported error numbers *)
let handleParseError lexbuf =
  let f, l, c, d = localize lexbuf in
    if !emacsStyle then (
      Format.eprintf "File \"%s\", line %i, characters %i-%i:\n" f l c d;
      Format.eprintf "Syntax Error\n@?"
    ) else (
      prerr_string "Line "; prerr_int l;
      prerr_string ":Par.Err("; prerr_int !errorNumber; prerr_string "): "
    );

  (* THIS CODE SHOULD BE UPDATED TO REFLECT ERRORS OF THE IF2ATSE *)
    (match !errorNumber with
       |  1 -> prerr_endline "Bad declaration for operators."
       |  2 -> prerr_endline "Bad declaration for operators."
       |  3 -> prerr_endline "Variables label begins with a uppercase."
       |  4 -> prerr_endline "Bad declaration for variables."
       |  5 -> prerr_endline "The TRS MUST be nammed."
       |  6 -> prerr_endline "The symbol: is the separator for associating an arity to a symbol."
       |  7 -> prerr_endline "Missing left parenthesis."
       |  8 -> prerr_endline "Missing right parenthesis."
       |  9 -> prerr_endline "Missing comma or right parenthesis."
       | 10 -> prerr_endline "Missing identifier for automaton."
       | 11 -> prerr_endline "Missing states declaration."
       | 12 -> prerr_endline "Either bad separator (' ') or empty states list."
       | 13 -> prerr_endline "Missing final states declaration."
       | 14 -> prerr_endline "Error in transition writing."
       | 15 -> prerr_endline "Not a normalized transition."
       | 16 -> prerr_endline "Missing symbol -> or not expected form in the lhs of the transition."
       | 17 -> prerr_endline "The rhs of the transition is not a state."
       | 18 -> prerr_endline "Missing comma or right parenthesis."
       | 19 -> prerr_endline "Missing left parenthesis."
       | 20 -> prerr_endline "Missing identifier for the approximation."
       | 21 -> prerr_endline "Missing key word 'Rules'."
       | 22 -> prerr_endline "Missing Coma or right parenthesis."
       | 23 -> prerr_endline "Either empty approximation or missing '['."
       | 24 -> prerr_endline "Error in transition writing."
       | 25 -> prerr_endline "Missing -> in the LHS of the approximation rule."
       | 26 -> prerr_endline "A RHS of a transition is neither a variable nor a state."
       | 27 -> prerr_endline "Missing ']'."
       | 28 -> prerr_endline "Missing '->': approx rule is of the form [...] -> [...]."
       | 29 -> prerr_endline "Missing '[' in the RHS of the approximation rule."
       | 30 -> prerr_endline "Missing ']' in the RHS of the approximation rule."
       | 31 -> prerr_endline "Transitions must be of the form f(t1,...,tn) -> tn+1 where ti is either a variable or a state."
       | 32 -> prerr_endline "Missing '->' for a transition declaration."
       | 33 -> prerr_endline "Missing coma within a transition or ')'."
       | 34 -> prerr_endline "State label MUST begin with the lower case 'q'."
       | 35 -> prerr_endline "An integer is expected."
       | 36 -> prerr_endline "Functionnal symbols begin with a case different of 'q'. Maybe the separator is not the one expected, i.e., ' ', /\\,','."
       | 37 -> prerr_endline "Variables label begins with a uppercase."
       | 38 -> prerr_endline "State in a rule is forbidden."
       | 39 -> prerr_endline "Variables are not allowed in automaton transitions."
       | 40 -> prerr_endline "Variable cannot be used as a functionnal symbol."
       | 41 -> prerr_endline "An identifier is required for labelling the set of terms."
       | 42 -> prerr_endline "Term is not of the expected form."
       | 43 -> prerr_endline "One term is required in a set at least."
       | 44 -> prerr_endline "The arity of a functionnal symbol does not fit with the signature."
       | 45 -> prerr_endline "An atomic data declaration is missing in the signature."
       | 46 -> prerr_endline "A variable occuring at this line is missing in the declaration section."
       | 47 -> prerr_endline "A state is unknown."
       | 48 -> prerr_endline "A final state has not been declared."
       | 49 -> prerr_endline "Missing an identifier for the set of equations."
       | 50 -> prerr_endline "An equation is of the form t1 = t2 (if conditions)."
       | 51 -> prerr_endline "Silent variable are not allowed at this placeS."
       | 52 -> prerr_endline "One condition is required at least. Conditions are of the form x <-> y where x and y are variables"
       | 53 -> prerr_endline "/\\ is expected between two conditions."
       | 54 -> prerr_endline "Right parenthesis is missing."
       | 55 -> prerr_endline "A condition is not well-formed i.e. of the form x <-> y where x and y are variables."
       | 56 -> prerr_endline "A condition for an equation is of the form Variable != Term and the left hand side of one of your conditions is not declared as a variable."
       | 57 -> prerr_endline "A cunjunction of conditions is required."
       | 58 -> prerr_endline "For now only equations are available for designing approximations."
       | 59 -> prerr_endline "Left (right) hand side of an equation must be different of _."
       | 60 -> prerr_endline "A variable is expected as left hand side of an equation condition."
       | 61 -> prerr_endline "A contextual equation is of the form: [pattr1, pattr2...,pattrn] => t1 = t2 (if conditions)."
         
       | 62 -> prerr_endline "A variable is used as a functionnal symbol."
       | 63 -> prerr_endline "A functionnal symbol declaration is missing."
       | 64 -> prerr_endline "Variables are forbidden in an initial set of terms."
       | 65 -> prerr_endline "Conditions are of the form x <-> y where x and y are variables."
       | 66 -> prerr_endline "Invalid equation."
       | 67 -> prerr_endline "Invalid equation list."
       | 68 -> prerr_endline "Declaration of a Special State forbidden."
       |  _ -> prerr_endline ("No description (err no "^ (string_of_int !errorNumber) ^") .")
    );
    flush stderr;
    exit (-1)

