
open Global
open Term
open Substitution
open Definitions

(* Conditionnal term rewriting system *)

type condition = 
  | Equal of (term * term)
  | Joinable of (term * term)


exception IncriminatedCondition of condition


let pcondition_to_condition = 
  let translate (t1, t2) = 
    (Term.pTerm_to_term t1, Term.pTerm_to_term t2)
  in
    function
      | PEqual cond -> Equal (translate cond) 
      | PJoinable cond -> Joinable (translate cond)
      | _ -> failwith "Rewriting:pcondition_to_condition: not covered yet"


type rule = (term * term * condition list)
type trs = rule list

exception InclusionTestFailed of string
exception NotHandledYet of string
exception RecapError of string

let rule_to_string ((lhs, rhs, cond): rule): string = 
  (Term.to_string lhs)^" -> "^(Term.to_string rhs)^
    (if cond = [] then "" else " if ") ^
    (String.concat " /\\ " 
       (List.map (function 
		    | Joinable(t1,t2) -> 
		        (Term.to_string t1) ^ "<->" ^ (Term.to_string t2)
		    | _ -> failwith "Rewriting:rule_to_string: Not supported yet"
                 ) cond
       )
    )

let to_string (rules :trs) : string = 
  String.concat "\n" (List.map rule_to_string rules)


(*
  The following function tests for a rule lhs -> rhs whether 
  var(rhs) is a subset of var(lhs) or not.
*)
let variables_inclusion_test (lhs: term) (rhs: term) : bool = 
  List2.subseteq (Term.vars rhs) (Term.vars lhs)


let vars ((l,r,c): rule): variable list = Term.vars l

let vars_trs (trs: rule list): variable list =
  List.fold_right (fun r -> List2.union (vars r)) trs []


let rec variables_inclusion_test_for_cond new_conds vars_l = 
  match new_conds with 
    | [] -> ()
    | Equal (t1, t2)::rem -> 
	if List.for_all (fun x -> List.mem x vars_l) ((Term.vars t1) @ (Term.vars t2)) then 
	  variables_inclusion_test_for_cond rem vars_l
	else
	  raise (IncriminatedCondition (Equal (t1, t2)))

    | Joinable (t1,t2)::rem -> 
	if List.for_all (fun x -> List.mem x vars_l) ((Term.vars t1) @ (Term.vars t2)) then
	  variables_inclusion_test_for_cond rem vars_l
	else
	  raise (IncriminatedCondition (Joinable(t1, t2)))


(* Is a TRS shallow *)

let rule_shallow ((_, r ,_): rule) : bool =
  Term.height r <= 1

let shallow (rules: trs): bool =
  List.for_all rule_shallow rules

(*
The following function returns a trs from the parsed data 
only if the parsed rules are of the expected form.

 *)


let rec trs_from_parser_format (rulel: pRule list): trs = 
  match rulel with 
    | [] -> [] 
    | Usual (lhs, rhs)::rem -> 
	let nlhs = Term.pTerm_to_term lhs and nrhs = Term.pTerm_to_term rhs in 
	  if variables_inclusion_test nlhs nrhs then 
	    try
	      Term.non_linearity_checking nlhs;
	      (nlhs, nrhs, [])::(trs_from_parser_format rem)
	    with
		Term.Non_linear -> 
                  raise (RecapError ("The rule "^(Term.to_string nlhs)^" -> "^(Term.to_string nrhs)^" is not left-linear."))
	  else
	    raise (InclusionTestFailed ("Vars("^(PTerm.to_string rhs)^") is not a subset of Vars("^(PTerm.to_string lhs)^")"))
    | Conditionnal (lhs, rhs, conds) :: rem -> 
	let nlhs = Term.pTerm_to_term lhs and nrhs = Term.pTerm_to_term rhs in 
	  if variables_inclusion_test nlhs nrhs then 
	    let new_conds = List.map pcondition_to_condition conds in
	      try
		variables_inclusion_test_for_cond new_conds (Term.vars nlhs);
		Term.non_linearity_checking nlhs;
		(nlhs, nrhs, new_conds)::(trs_from_parser_format rem);
	      with 
		| IncriminatedCondition (Equal (t1, t2)) -> 
                    raise (InclusionTestFailed ("Some variables of the condition equal("^(Term.to_string t1)^","^(Term.to_string t2)^") of the rule "^(PTerm.to_string rhs)^" -> "^(PTerm.to_string lhs)^" are not included in vars("^(PTerm.to_string rhs)^")"))
		| IncriminatedCondition (Joinable (t1, t2)) ->
                    raise (InclusionTestFailed ("Some variables of the condition "^(Term.to_string t1)^"<->"^(Term.to_string t2)^" of the rule "^(PTerm.to_string rhs)^" -> "^(PTerm.to_string lhs)^" are not included in vars("^(PTerm.to_string rhs)^")"))
		| Term.Non_linear ->
                    raise (RecapError ("The rule "^(Term.to_string nlhs)^" -> "^(Term.to_string nrhs)^" is not left-linear."))
	  else
	    raise (InclusionTestFailed ("Vars("^(PTerm.to_string rhs)^") is not a subset of Vars("^(PTerm.to_string lhs)^")"))
   
(*raise (NotHandledYet "Condionnal rules are not yet handled.")*)







(* Application of a substitution *)

let rec subst_apply (s: substitution) : term -> term = function
  | Fsym (f, largs) -> Fsym (f, List.map (subst_apply s) largs)
  | Var x -> 
      (try Substitution.find x s with
	 | Not_found -> Var x
      )
  | State q -> State q

let substset_apply (es: substitutionset) (t: term): (term * Formula.t) list =
  SubstitutionSet.fold (fun s f acc -> (subst_apply s t, f)::acc) es []

      
