open Definitions
open Term
(* open Set2 *)


(** A substitution associates a variable to an element of the range *)
module Substitution =
struct

  module Subtemp = Map.Make(Variable) (* Module pour les applications à argument Variable et à valeurs indéfinies *)
  type substitution = term Subtemp.t (* on particularise : applications à argument Variable et à valeurs termes *)

  type t = substitution

  let empty = Subtemp.empty
  let add = Subtemp.add
  let equal = Subtemp.equal
  let find = Subtemp.find
  let fold = Subtemp.fold
  let map = Subtemp.map
  let compare (s1: substitution) (s2: substitution) =
    (Subtemp.compare Variable.compare s1 s2)

  (* Cardinal of a substitution *)
  (* B: Mouais c'est pas tres mathematique comme def... *)
  let cardinal s =
    Subtemp.fold (fun _ _ i -> i + 1) s 0


  (* s1 and s2 are NO LONGER supposed to be two substitutions with disjoint domains *)
  let union s1 s2 =
    Subtemp.fold Subtemp.add
      (* fun x t subst -> if Subtemp.mem x subst then subst else
	 Subtemp.add x t subst *)
      s1 s2

  let to_string (s: substitution) =
    let str =
      Subtemp.fold
        (fun x t str ->
           ", " ^ (Variable.to_string x) ^ " -> " ^ (Term.to_string t) ^ str
        ) s ""
    in
      "{" ^ (if str = "" then "" else Str.string_after str 2) ^ "}"
        (*
          let rec aux l = function
          | [] -> ""
          | [(x, v)] -> (Variable.to_string x) ^ " -> " ^ (Term.to_string v)
          | (x, v)::rem -> (Variable.to_string x) ^ " -> "^ (Term.to_string v) ^ ", " ^ (aux rem)
          in
          let res = ref [] in
          let _ = Subtemp.iter (fun x v -> res:= (x,v)::!res) s in
          "{"^(aux !res)^"}"
        *)


	(** Quel est le sens de cette exception ?*)
  exception Inclusion

  (** [covers s1 s2] lorsque le domaine de [s1] est inclus dans celui de [s2]
  et pour tout x du domaine de [s1], s1(x)=s2(x) ;
	ie [s1] est plus général que [s2]*)
	(* if value for all keys is the same *)
  let covers (s1: substitution) (s2: substitution) : bool =
    try
      Subtemp.fold
        (fun key t b ->
	   if key = "_" then
             failwith "No _ in substitution at that level"
	   else if t = Subtemp.find key s2 then
             b
           else
	     raise Inclusion
        )
	s1 true
    with Not_found | Inclusion -> false

  (* Substitution inclusion *)

  (* fold with exception, in the case where an association is not included
     in s2, in order not to follow all s1... due to unfriendly interface
     of maps: no for_all for instance... *)
  (* Note that variable "_" plays here a specific role.
     if _ is associated to v1 on one side it may be associated (or not)
     on the other side *)

  let is_included (s1: substitution) (s2: substitution) : bool =
    try
      Subtemp.fold
        (fun key t b ->
	   if key = "_" then
             b
	   else if t = Subtemp.find key s2 then
             b
           else
	     raise Inclusion
        )
	s1 true
    with Not_found | Inclusion -> false

  (* Extended substitution equivalence such that {X -> q1, Z -> q3} is equivalent in
     {Y -> q2, Z-> q3}, i.e. substitution are always equivalent on variables they do not
     share.  *)

  (* Used for matching lhs and rhs of equations. Those equations only produce Trivial
     formulas in substitution, no check has to be performed on formulas *)
  let extended_equivalence (s1: substitution) (s2: substitution) : bool =
    try
      Subtemp.fold
        (fun key t b ->
	   if key = "_" then
             b
	   else
             try
	       if t = Subtemp.find key s2 then
                 b
               else
                 raise Inclusion
	     with Not_found -> b
        )
	s1 true
    with Inclusion -> false

end (* Substitution *)


type substitution = Substitution.t


(* A substitutionPhi contains a substitution and a formula *)
module SubstitutionPhi = struct

  type t = { value : substitution;
             phi: Formula.t
           }

  exception Inclusion


  let empty =
    { value = Substitution.empty; phi = Formula.mk_false }

  let add x t s =
    { s with
       value = Substitution.add x t s.value
    }

  let define sigma phi =
    { value = sigma; phi = phi }

  let addPhi p s =
    { s with
        phi = Formula.mk_or s.phi p
    }

  let addBoth x t f s =  (* AddPhi f (add x t s) *)
    { value = Substitution.add x t s.value;
      phi = Formula.mk_or s.phi f
    }

  let compare (s1: t) (s2: t) =
    Substitution.compare s1.value s2.value

  (* Cardinal of a substitution *)
  let cardinal s = Substitution.cardinal s.value

  (* s1 and s2 are NO LONGER supposed to be two substitutions with disjoint domains *)
  (* Union of maps and AND of formulas *)
  let union s1 s2 =
    { value = Substitution.union s1.value s2.value;
      phi = Formula.mk_and s1.phi s2.phi
    }

  let to_string (s: t) =
    (Substitution.to_string s.value) ^ "[" ^ (Formula.to_string s.phi) ^ "]"

  (* coversValue : s1 covers by s2 if value is the same *)
  let coversValue (s1: t) (s2: t) : bool =
    Substitution.covers s1.value s2.value


  (* Covered substitution: s1 covers by s2 if value is the same AND
     !!!formula of s2 implies s1!!!

     Indeed if s2=sigma||A
               s1=sigma||A\/B

     s1 covers s2 because it is more general (more application cases)
  *)

  let covers (s1: t) (s2: t) : bool =
    coversValue s1 s2 && (Formula.subsume s2.phi s1.phi)

  let equal (s1: t) (s2: t) : bool =
    covers s1 s2 && (covers s2 s1)

  (* Substitution inclusion *)

  (* fold with exception, in the case where an association is not included
     in s2, in order not to follow all s1... due to unfriendly interface
     of maps: no for_all for instance... *)
  (* Note that variable "_" plays here a specific role.
     if _ is associated to v1 on one side it may be associated (or not)
     on the other side *)

  (* Does not take formulas into account! *)
  let is_included (s1: t) (s2: t) : bool =
    Substitution.is_included s1.value s2.value


  (* Extended substitution equivalence such that {X -> q1, Z -> q3} is equivalent in
     {Y -> q2, Z-> q3}, i.e. substitution are always equivalent on variables they do not
     share.  *)

  (* Used for matching lhs and rhs of equations. Those equations only produce Trivial
     formulas in substitution, no check has to be performed on formulas *)

  let extended_equivalence (s1: t) (s2: t) : bool =
    Substitution.extended_equivalence s1.value s2.value
end

type substitutionPhi = SubstitutionPhi.t


(* A substitution set is a map associating a substitution (Subtemp.t) to a formula *)
module SubstitutionSet =
struct
  module Substsettemp = Map.Make(Substitution)

  open SubstitutionPhi

  (* open SubstitutionPhi *)
  (* exception Not_found= Substsettemp.Not_found *)
  type substitutionset = Formula.t Substsettemp.t (* application Substitution → Formula *)

  type t = substitutionset
  let empty = Substsettemp.empty

  let to_list es =
    Substsettemp.fold
      (fun s phi l -> (SubstitutionPhi.define s phi)::l) es []


  let to_string es =
    "{" ^ (String.concat "," (List.map SubstitutionPhi.to_string (to_list es)) ) ^ "}"

  let list_to_string les=
    "[" ^ (String.concat ";" (List.map to_string les)) ^"]"


  (* Membership in set of substitution is replaced by coverage *)
  (** A substitution [s1] is in [es] if [s1.value] is in [es] and the formula associated
     in to [s1.value] in [es] is IMPLIED by [s1.phi]
		
		noté [s1 ⋹ es]
		
		*)
  let mem s1 es =
    try
      let phi2 = Substsettemp.find s1.value es in
	Formula.subsume s1.phi phi2
    with
      | Not_found -> false


  (*let replace s phi es =
    Substsettemp.add s phi (Substsettemp.remove s es)
		(* étape remove inutile ; en effet selon la doc de add, la liaison précédente est effacée *)
	*)
	let replace = Substsettemp.add;;

  (* Addition of substitution to a set of substs, checks if there is a similar subst
     in the set. In case, adds a disjonction to the formula *)
  let add (s: substitutionPhi) es =
    try
      let phi2 = Substsettemp.find s.value es in
	if Formula.subsume phi2 s.phi then
          es
	else
          replace s.value (Formula.mk_or phi2 s.phi) es
    with
      | Not_found -> Substsettemp.add s.value s.phi es


  let fold = Substsettemp.fold

  let addPhi_to_all (f: Formula.t) (es: substitutionset) : substitutionset =
    fold (fun s f2 es -> Substsettemp.add s (Formula.mk_and f2 f) es) es empty

  let from_list l = List.fold_right (fun s es -> add s es) l empty


  (* Substitution set difference
     in diff A B, we remove a substitution s associated to formula fB in B
     from substitutions of A if s occurs in A and is associated to fA such
     that fA => fB.
  *)
  let diff (esA: substitutionset) (esB: substitutionset) =
    fold
      (fun s fA es ->
         try
           let fB = Substsettemp.find s esB in
             if Formula.subsume fA fB then
               es
             else
               Substsettemp.add s fA es
         with
           | Not_found -> Substsettemp.add s fA es
      ) esA empty


  let is_empty =  Substsettemp.is_empty

  (* B: Assez cher : 3 * |es|
  *)
  let filter f es =
     from_list (List.filter f (to_list es))
       (* B: Autre proposition (ne type pas plus bas... )
         fold
         (fun sigma phi acc ->
         if f sigma then Substsettemp.add sigma phi es else es)
         es Substsettemp.empty
       *)



  (* Semantic inclusion of substitution set. A set es1 is included in es2 if
     all substitutions of es1 are present (syntactically) in es2 and with EQUIVALENT formulas.
     Only a criterion if Formula.subsume is a criterion.
  *)

  let included es1 es2 =
    Substsettemp.fold
      (fun s p b ->
	 try
           Formula.equival p (Substsettemp.find s es2) && b
	 with Not_found -> false
      ) es1 true


  (* Semantic equivalence of substitution sets, see included  *)
  let equal es1 es2 = included es1 es2 && (included es2 es1)

  (* let singleton s = add s empty
	(* le module Map contient une fonction singleton depuis OCaml 3.12*)
	 *)
	let singleton (s: substitutionPhi) : substitutionset = Substsettemp.singleton s.value s.phi;; 

  let
elements s : substitutionPhi list = to_list s
    (*
       Substsettemp.fold (fun e phi l -> {value=e;phi=phi}::l) s []
    *)
  let union es1 es2=
    Substsettemp.fold
      (fun e phi es -> add {value=e;phi=phi} es) es1 es2

  (* Combination of substitionset list *)
  (* combine of [{{x -> q1},{x -> q2}};{{y->q3,z->q4}}] =
     {{x -> q1,y->q3,z->q4},{x -> q2,y->q3,z->q4}}
     if one set of susbt is empty, so is the combination.
     Combination of {({x->q1},Phi1), ({x->q2},Phi2)} with
  {({y -> q3},Phi3), ({y -> q4},Phi4)} results in
  {({x -> q1, y -> q3},Phi1 et Phi3), ({x->q1,y->q4},Phi1 et Phi4) etc. *)

  let rec combine (sl: substitutionset list): substitutionset=
    let rec one_subst_combine (s1: substitutionPhi) (s2: substitutionPhi list): substitutionset=
      match s2 with
	  [] -> Substsettemp.empty
	| s::rem ->
	    add (SubstitutionPhi.union s s1)(one_subst_combine s1 rem)
    in
    let rec substens_combine (s1: substitutionPhi list) (s2: substitutionPhi list)=
      match s1 with
	  [] -> Substsettemp.empty
	| e::rem ->
	    let combine_first= (one_subst_combine e s2) in
	      union combine_first (substens_combine rem s2)
    in
      match sl with
	  [] -> Substsettemp.empty
	| [s] -> s
	| s::rem ->
	    if Substsettemp.is_empty s then Substsettemp.empty
	    else
	      let res_rem= combine rem in
		if Substsettemp.is_empty res_rem
		then Substsettemp.empty
		else substens_combine (to_list s) (to_list res_rem)


(* Matching substitution checks if at least one substitution of es1 is included in
   one substitution of es2 (or one of es2 is included in one of es1)
   Note also that substitutions are not compared on variables not occurring in
   both es1 and es2. For instance, {{X -> q1}} and {{Y -> q2}} is matching. This
   is due to the fact that it is necessary for equations of the form s(X)=s(Y) where
   any substitution on X and Y is valid.
*)

(* Matching substitution w.r.t. extended_equivalence, no check on formulas since
   it is used for equations (trivial formulas on subst produced for equations) *)

let matching_substitutionset (es1: substitutionset) (es2: substitutionset): bool=
  if is_empty es1 then false
  else
    if is_empty es2 then false
    else
      List.exists
	(function elt -> List.exists
	   (function elt2 -> SubstitutionPhi.extended_equivalence elt elt2) (to_list es2)) (to_list es1)

(* Intersection of substitution sets where identity of substitutions is replaced by
   extended_equivalence. For extended equations -> no need to check formulas associated
   to substitution they are all trivial. *)

  let extended_intersection (es1: substitutionset) (es2: substitutionset): substitutionset=
    let rec aux sl1 sl2 sres=
      match sl1 with
	  [] -> sres
	| s1::rem -> (* Solutions are union of substitutions of es1 and es2 such that
			they are identitcal w.r.t. extended_equivalence *)
	    let newsres= List.fold_right
			   (function s2-> function es ->
			      if SubstitutionPhi.extended_equivalence s1 s2
			      then (add (SubstitutionPhi.union s1 s2) es)
			      else es)
			   sl2 sres
	    in
	      aux rem sl2 newsres
    in
      aux (to_list es1) (to_list es2) empty

end

type substitutionset= SubstitutionSet.t


module SubstitutionTable=
struct
  (* in implementation of substable set of substitutions are ordered by redex
     state then by rule state. But presentation (with to_string for instance) is
     just the opposite. The reason for ordering table by redex state firt is
     that merge applied on redex states merges substitutions for rule states. It
     is thus important to be abble to find all subtables for 2 distinct redex
     states in order to merge them *)
  (* to 2 states we associate a s

     qredex -> qlhs -> substitution set where each substitution is associated to
     a formula Phi. Formula Phi tells which are the q1->q2 necessary for the
     substitution to exists.  *)

  type substable= (state, (state, substitutionset) Hashtbl.t) Hashtbl.t
  type t = substable
  let new_substable () : substable = Hashtbl.create Common.max_state_symb
  exception Couple_not_found

  (* Finding a substitution s in a table t for a given couple of states cq.
     Throws Couple_not_found if cq is not in the table  *)

  let find ((qlhs,qred): state * state) (t: substable): substitutionset=
    try Hashtbl.find (Hashtbl.find t qred) qlhs
    with Not_found -> raise Couple_not_found

  (* Checks if a substitution is COVERED by a table *)
	(** [mem cq s t] renvoie [s ⋹ t(cq.fst, cq.snd)] *) 
  let mem (cq: state * state) (s: substitutionPhi) (t: substable): bool=
    try (let es= (find cq t) in SubstitutionSet.mem s es)
    with Couple_not_found -> false


(* Adding a couple with an empty substitution for marking purpose*)
(* The couple is supposed to be non present in the table *)
  let mark ((qlhs,qred): state * state) (t: substable): unit=
    try
      (let sub_by_qlhs= Hashtbl.find t qred in
	 try
	   (let _= Hashtbl.find sub_by_qlhs qlhs in 
            failwith ("The state (qlhs,qred) is already present in the table!")
           )
         with Not_found -> Hashtbl.add sub_by_qlhs qlhs SubstitutionSet.empty
      )
    with Not_found ->
      let newsub_by_qlhs= Hashtbl.create Common.max_rule_node in
	begin
	  Hashtbl.add newsub_by_qlhs qlhs SubstitutionSet.empty;
	  Hashtbl.add t qred newsub_by_qlhs
        end

(* Adding of a substitution s in a table t for a given couple of states cq.
   Returns true if the substitution was not *COVERED* by the table *)

  let add (s: substitutionPhi) ((qlhs,qred): state * state) (t: substable): bool=
    try
      (let sub_by_qlhs= Hashtbl.find t qred in
	 try
	   (let ens_subst= Hashtbl.find sub_by_qlhs qlhs in
	      if SubstitutionSet.mem s ens_subst then false
	      else
		begin
		  Hashtbl.replace sub_by_qlhs qlhs (SubstitutionSet.add s ens_subst);
		  true
		end)
	 with Not_found ->
	   begin
	     Hashtbl.add sub_by_qlhs qlhs (SubstitutionSet.singleton s);
	     true
	   end)
    with Not_found ->
      let newsub_by_qlhs= Hashtbl.create Common.max_rule_node in
	begin
	  Hashtbl.add newsub_by_qlhs qlhs (SubstitutionSet.singleton s);
	  Hashtbl.add t qred newsub_by_qlhs;
	true
      end


  (* Adding a set of substitution es in a table t for a given couple of
     states cq. Returns the set of substitution that were not already *COVERED* by
     the table, and the empty set otherwise (in the case where they are
     all in the table. The first state is for lhs, the second one for redex *)

  let add_set (es: substitutionset) (cq: state * state) (t: substable): substitutionset=
    SubstitutionSet.filter (function s -> add s cq t) es


let to_string (t: substable): string=
  let all_sub1= Hashtbl2.elements t in
  let all_sub= List.map (function (q,symtable) -> (q, (Hashtbl2.elements symtable))) all_sub1 in
  let rec to_string_aux1 l q=
    match l with
	[] -> ""
      | (q2,es)::rem -> "("^(State.to_string q2)^","^(State.to_string q)^") -> "^
			(SubstitutionSet.to_string es)^"\n"^(to_string_aux1 rem q) in
  let rec to_string_aux l=
    match l with
	[] -> ""
      | (q,l)::rem -> (to_string_aux1 l q)^(to_string_aux rem)
  in
    to_string_aux all_sub


let list_to_string (l: ((state * state) * substitutionset) list): string=
  List.fold_right (function ((q1,q2),es) -> function s ->
		     ("("^(State.to_string q1)^","^(State.to_string q2)^") -> "^
			(SubstitutionSet.to_string es)^"\n"^s))
    l ""


  let copy (t: substable): substable=
    Hashtbl2.copy_level2 t

end

type substable= SubstitutionTable.t




