open Definitions

module Statetable : Set2.S with type elt = state list

type table = Statetable.t
type symtable = (symbol, table) Hashtbl.t
type transtable = (state, symtable) Hashtbl.t
type satable =
    (state, (symbol, (state, table) Hashtbl.t) Hashtbl.t) Hashtbl.t


type ftable = (symbol, Stateset.t) Hashtbl.t
type infered = Away | Alive of int | Ghost
type imatrix = infered array array
type fmatrix = Formula.t array array

class virtual epsilontbl : int -> object ('a)
    method copy : 'a
    method resize : int -> unit

    method last : int
    method incr_last : unit
		method add_to_last : int -> unit

    method fclosure : fmatrix
    method private add : state -> Stateset.elt -> unit
    method private remove : state -> Stateset.elt -> unit
    method private virtual rm_atom : state -> state -> unit


    method succs : (state, Stateset.t) Hashtbl.t
    method preds : (state, Stateset.t) Hashtbl.t

    method succs_of : state -> Stateset.t
    method preds_of : state -> Stateset.t

end

class epsilonRtbl : int -> object ('a)
  method copy : 'a
  method resize : int -> unit

  method last : int
  method incr_last : unit
	method add_to_last : int -> unit

  method fclosure : fmatrix
  method private add : state -> Stateset.elt -> unit
  method private remove : state -> Stateset.elt -> unit
  method private rm_atom : state -> state -> unit

  method addE : state -> state -> unit
  method addR : state -> Stateset.elt -> Formula.t -> unit
  method removeE : state -> state -> unit


  method succs : (state, Stateset.t) Hashtbl.t
  method preds : (state, Stateset.t) Hashtbl.t

  method succs_of : state -> Stateset.t
  method preds_of : state -> Stateset.t
end


class epsilonEtbl : int -> object ('a)
  method copy : 'a
  method resize : int -> unit

  method degrees : imatrix
  method fclosure : fmatrix
  method private add : state -> state -> unit
  method private remove : state -> state -> unit
  method private rm_atom : state -> state -> unit

  method last : int
  method incr_last : unit
	method add_to_last : int -> unit

  method addE :  ?degree:int -> state -> state -> unit
  method removeE : state -> state -> unit

  method succs : (state, Stateset.t) Hashtbl.t
  method preds : (state, Stateset.t) Hashtbl.t

  method succs_of : state -> Stateset.t
  method preds_of : state -> Stateset.t
end

type t = {
  transitions : transtable;
  satable : satable;
  ftable : ftable;
  epsilonR : epsilonRtbl;
  epsilonE : epsilonEtbl;
  mutable next_state : state;
}

type external_symtrans = symbol * state list * state
type external_epsilon = state * state * Formula.t
(* En sursis... *)
type external_trans =
  | Trans of external_symtrans
  | Epsilon of external_epsilon



(* nouvel ensemble de transitions vide *)
val make : unit -> t
(* realise une copie profonde de delta *)
val copy : t -> t

(* obtenir un nouvel etat *)
val make_state : t -> state
val make_n_states : t -> int -> unit

(* Set the maximum value of state counter. Used for building a TA
   from a list of external_symtrans (no epsi) *)
val set_state : t -> state -> unit



val add : t -> external_symtrans -> unit
(* addR et addE ajoute une transition epsilon E ou R
   et maintiennent directement la cloture
*)
val addE : t -> state -> state -> unit
val addR : t -> external_epsilon -> unit
(* Les memes fonctions mais retourne vrai ssi la transition a bien ete ajoutee *)
val add' : t -> external_symtrans -> bool
val addE' : t -> state -> state -> bool
val addR' : t -> external_epsilon -> bool

(* Operation de raffinement: prend une formule et elague l'automate en consequence... *)
val prune : t -> Formula.t -> unit


(* [is_succRE Δ q1 q2] <=> q2 ———>^*_RE q1 *)
val is_succRE : t -> state -> state -> bool

(* same but returns the formula phi such that q2 -> q1 *)
val is_succRE_formula : t -> state -> state -> Formula.t

(* [is_succE Δ q1 q2] <=> q2 ———>^*_E q1 *)
val is_succE : t -> state -> state -> bool

(* (q', α) ∈ [get_predRE Δ q] <=> q' ––α—>^*_RE* q *)
val get_predRE : t -> state -> (state * Formula.t) list

(* (q', α) ∈ [get_succRE Δ q] <=> q ––α—>^*_RE q' *)
val get_succRE : t -> state -> (state * Formula.t) list

(* q' ∈ [get_succE Δ q] <=> q –––—>^+_E q' *)
val get_succE : t -> state -> state list

(* q' ∈ [get_predE Δ q] <=> q' –––—>^+_E q *)
val get_predE : t -> state -> state list

(* En sursis aussi... *)
val get_all_states : t -> state list

(* Liste des composantes fortement connexes de E *)
val get_strongly : (state -> state -> bool) -> t -> state list list

val to_string : t -> string

val trans_to_string : transtable -> string
val epsilonR_to_string : ?timbuk:bool -> t -> string


val fmatrix_to_string :
  < fclosure : Formula.formula array array; last : int; .. > -> string
val imatrix_to_string :
  ?ghost:bool ->
  ?timbuk:bool -> < degrees : infered array array; last : int; .. > -> string

val to_timbuk : t -> string


val print_fclosure : < fclosure : Formula.t array array; last : int; .. > -> unit

(* Ca aussi en sursis... *)
val trans_to_transitionsBySymbols :
  transtable -> (symbol, (Statelist.t * state) list) Hashtbl.t

val not_epsi : external_trans -> bool
val ext2trans : external_trans -> external_symtrans
val ext2epsi : external_trans -> external_epsilon
val partition : external_trans list -> external_symtrans list * external_epsilon list
val trans_to_list : transtable -> external_trans list
val to_list : t -> external_trans list
val to_list_noE : t -> external_trans list
val list_to_string : external_trans list -> string
val external_to_string : external_trans -> string
val epsiext_to_string : external_epsilon -> string
val symextlist_to_string : external_symtrans list -> string
val symext_to_string : external_symtrans -> string
val epsiextlist_to_string : external_epsilon list -> string
val get_all_predPhi : Stateset.t -> t -> (state * (state * Formula.t)) list
val get_all_pred : Stateset.t -> t -> Stateset.t
val get_trans : state -> symbol -> t -> external_symtrans list
val get_trans_sym_constant : symbol -> t -> external_symtrans list
val get_trans_sym_parameter : symbol -> state -> t -> external_symtrans list
val simple_get_all_trans : state list -> t -> external_symtrans list

(** [simple_get_trans f Δ] gets all transitions of Δ with top symbol f (not modulo epsilon) *)
val simple_get_trans_f : symbol -> t -> external_symtrans list
val constant_table : table -> bool


val propagate_epsilon :
  external_symtrans list -> state list -> t -> external_symtrans list
val propagate_epsilon_up :
  external_epsilon list -> external_symtrans list -> external_symtrans list

  (* B: Le partage des donnees de test entre differents fichiers...
     Bof bof bof... (T: Comment ça bof bof :-)... ouais bon c'est nul
     mais bon c'était économique au départ. )
     B: ouais, mais vue la depense de ligne de codes produites par copier coller dans les
     tests, je pense qu'on est plus a 5 lignes, et ces declarations ne sont utilisees qu'a
     un seul endroit :-p
     Pas urgent mais a revoir !
  *)


